# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import os
import h5py
import numpy as np
import warnings

from deoxys.keras.callbacks import CSVLogger
from ..model.callbacks import DeoxysModelCheckpoint, PredictionCheckpoint, \
    DBLogger
from ..model import model_from_full_config, model_from_config, load_model
from deoxys_vis import plot_log_performance_from_csv, mask_prediction, \
    plot_images_w_predictions, read_csv
from ..database import Tables, ExperimentAttr, HDF5Attr, SessionAttr, \
    SessionStatus


class Experiment:
    MODEL_PATH = '/model'
    MODEL_NAME = '/model.{epoch:03d}.h5'
    BEST_MODEL_PATH = '/best'
    PREDICTION_PATH = '/prediction'
    PREDICTION_NAME = '/prediction.{epoch:03d}.h5'
    LOG_FILE = '/logs.csv'
    PERFORMANCE_PATH = '/performance'
    PREDICTED_IMAGE_PATH = '/images'
    TEST_OUTPUT_PATH = '/test'
    PREDICT_TEST_NAME = '/prediction_test.h5'

    def __init__(self,
                 log_base_path='logs',
                 best_model_monitors='val_loss',
                 best_model_modes='auto'):

        self.model = None

        self.architecture = None
        self.input_params = None
        self.model_params = None
        self.train_params = None
        self.data_reader = None
        self.weights_file = None

        self.log_base_path = log_base_path

        self.best_model_monitors = best_model_monitors if type(
            best_model_monitors) == list else [best_model_monitors]
        self.best_model_modes = best_model_modes if type(
            best_model_modes) == list else \
            [best_model_modes] * len(best_model_monitors)

        for i, (monitor, mode) in enumerate(zip(self.best_model_monitors,
                                                self.best_model_modes)):
            if mode not in ['auto', 'min', 'max']:
                warnings.warn('ModelCheckpoint mode %s is unknown, '
                              'fallback to auto mode.' % (mode),
                              RuntimeWarning)
                mode = 'auto'

            if mode == 'auto':
                if 'acc' in monitor or \
                        monitor.startswith('fmeasure') or \
                        'fbeta' in monitor:
                    self.best_model_modes[i] = 'max'
                else:
                    self.best_model_modes[i] = 'min'

    def from_full_config(self, file, weights_file=None, **kwargs):
        self.model = model_from_full_config(file,
                                            weights_file=weights_file,
                                            **kwargs)

        return self

    def from_config(self, architecture, input_params,
                    model_params, train_params,
                    dataset_params, weights_file=None):
        self.model = model_from_config(architecture, input_params,
                                       model_params, train_params,
                                       dataset_params,
                                       weights_file=weights_file)

        return self

    def from_file(self, filename):
        self.model = load_model(filename)

        return self

    def best_model(self):
        res = {}
        logger_path = self.log_base_path + self.LOG_FILE
        if os.path.isfile(logger_path):
            df = read_csv(logger_path, index_col='epoch',
                          usecols=['epoch'] + self.best_model_monitors)
            min_df = df.min()
            min_epoch = df.idxmin()
            max_df = df.max()
            max_epoch = df.idxmax()
            for monitor, mode in zip(self.best_model_monitors,
                                     self.best_model_modes):
                if mode == 'min':
                    val = min_df[monitor]
                    epoch = min_epoch[monitor]
                else:
                    val = max_df[monitor]
                    epoch = max_epoch[monitor]
                res[monitor] = {
                    'best': {
                        'val': val,
                        'epoch': epoch + 1
                    }}
        else:
            warnings.warn('No log files to check for best model')

        return res

    def run_experiment(self, train_history_log=True,
                       model_checkpoint_period=0,
                       prediction_checkpoint_period=0,
                       save_origin_images=False,
                       verbose=1,
                       epochs=None, initial_epoch=None, **custom_kwargs
                       ):
        log_base_path = self.log_base_path

        if self._check_run():
            if not os.path.exists(log_base_path):
                os.makedirs(log_base_path)

            kwargs = custom_kwargs or {}

            csv_logger_append = False

            if epochs:
                kwargs['epochs'] = epochs
            if initial_epoch:
                kwargs['initial_epoch'] = initial_epoch
                if initial_epoch > 0:
                    csv_logger_append = True

            callbacks = []

            if train_history_log:
                callback = self._create_logger(log_base_path,
                                               append=csv_logger_append)
                callbacks.append(callback)

            if model_checkpoint_period > 0:
                if not os.path.exists(log_base_path + self.MODEL_PATH):
                    os.makedirs(log_base_path + self.MODEL_PATH)

                callback = self._create_model_checkpoint(
                    log_base_path,
                    period=model_checkpoint_period)
                callbacks.append(callback)

            if prediction_checkpoint_period > 0:
                if not os.path.exists(log_base_path + self.PREDICTION_PATH):
                    os.makedirs(log_base_path + self.PREDICTION_PATH)

                callback = self._create_prediction_checkpoint(
                    log_base_path,
                    prediction_checkpoint_period,
                    use_original=save_origin_images
                )
                callbacks.append(callback)

            kwargs['callbacks'] = callbacks

            self.model.fit_train(**kwargs)

            return self

    def plot_performance(self):
        log_base_path = self.log_base_path

        if not os.path.exists(log_base_path + self.PERFORMANCE_PATH):
            os.makedirs(log_base_path + self.PERFORMANCE_PATH)

        if os.path.exists(log_base_path + self.LOG_FILE):
            print('\nPlotting performance metrics...')

            plot_log_performance_from_csv(
                filepath=log_base_path + self.LOG_FILE,
                output_path=log_base_path + self.PERFORMANCE_PATH)
        else:
            raise Warning('No log files for plotting performance')

        return self

    def plot_prediction(self, masked_images,
                        contour=True,
                        base_image_name='x',
                        truth_image_name='y',
                        predicted_image_title_name='Image {index:05d}',
                        img_name='{index:05d}.png'):
        log_base_path = self.log_base_path

        if os.path.exists(log_base_path + self.PREDICTION_PATH):
            print('\nCreating prediction images...')
            # mask images
            prediced_image_path = log_base_path + self.PREDICTED_IMAGE_PATH
            if not os.path.exists(prediced_image_path):
                os.makedirs(prediced_image_path)

            for filename in os.listdir(
                    log_base_path + self.PREDICTION_PATH):
                if filename.endswith(".h5") or filename.endswith(".hdf5"):
                    # Create a folder for storing result in that period
                    images_path = prediced_image_path + '/' + filename
                    if not os.path.exists(images_path):
                        os.makedirs(images_path)

                    self._plot_predicted_images(
                        data_path=log_base_path + self.PREDICTION_PATH
                        + '/' + filename,
                        out_path=images_path,
                        images=masked_images,
                        base_image_name=base_image_name,
                        truth_image_name=truth_image_name,
                        title=predicted_image_title_name,
                        contour=contour,
                        name=img_name)

        return self

    def run_test(self, use_best_model=False,
                 masked_images=None,
                 use_original_image=False,
                 contour=True,
                 base_image_name='x',
                 truth_image_name='y',
                 image_name='{index:05d}.png',
                 image_title_name='Image {index:05d}'):

        log_base_path = self.log_base_path

        test_path = log_base_path + self.TEST_OUTPUT_PATH

        if not os.path.exists(test_path):
            os.makedirs(test_path)

        if use_best_model:
            raise NotImplementedError
        else:
            score = self.model.evaluate_test(verbose=1)
            print(score)

            predicted = self.model.predict_test(verbose=1)

            # Create the h5 file
            filepath = test_path + self.PREDICT_TEST_NAME
            hf = h5py.File(filepath, 'w')
            hf.create_dataset('predicted', data=predicted)
            hf.close()

            if use_original_image:
                original_data = self.model.data_reader.original_test

                for key, val in original_data.items():
                    hf = h5py.File(filepath, 'a')
                    hf.create_dataset(key, data=val)
                    hf.close()
            else:
                # Create data from test_generator
                x = None
                y = None

                test_gen = self.model.data_reader.test_generator
                data_gen = test_gen.generate()

                for _ in range(test_gen.total_batch):
                    next_x, next_y = next(data_gen)
                    if x is None:
                        x = next_x
                        y = next_y
                    else:
                        x = np.concatenate((x, next_x))
                        y = np.concatenate((y, next_y))

                hf = h5py.File(filepath, 'a')
                hf.create_dataset('x', data=x)
                hf.create_dataset('y', data=y)
                hf.close()

            if masked_images:
                self._plot_predicted_images(
                    data_path=filepath,
                    out_path=test_path,
                    images=masked_images,
                    base_image_name=base_image_name,
                    truth_image_name=truth_image_name,
                    title=image_title_name,
                    contour=contour,
                    name=image_name)

        return self

    def run_lambda(self, lambda_fn, **kwargs):
        """
        Custom action between experiments
        """
        lambda_fn(self, **kwargs)
        return self

    def _create_logger(self, base_path, append=False):
        return CSVLogger(filename=base_path + self.LOG_FILE, append=append)

    def _create_model_checkpoint(self, base_path, period):
        return DeoxysModelCheckpoint(
            period=period,
            filepath=base_path + self.MODEL_PATH + self.MODEL_NAME)

    def _create_prediction_checkpoint(self, base_path, period, use_original):
        return PredictionCheckpoint(
            filepath=base_path + self.PREDICTION_PATH + self.PREDICTION_NAME,
            period=period, use_original=use_original)

    def _plot_predicted_images(self, data_path, out_path, images,
                               contour=True,
                               base_image_name='x',
                               truth_image_name='y',
                               title='Image {index:05d}',
                               name='{index:05d}.png'):
        hf = h5py.File(data_path, 'r')
        keys = list(hf.keys())
        for index in images:
            kwargs = {key: hf[key][index] for key in keys}
            img_name = out_path + '/' + name.format(index=index, **kwargs)
            if contour:
                mask_prediction(img_name,
                                image=kwargs[base_image_name],
                                true_mask=kwargs[truth_image_name],
                                pred_mask=kwargs['predicted'],
                                title=title.format(index=index, **kwargs))
            else:
                plot_images_w_predictions(
                    img_name,
                    image=kwargs[base_image_name],
                    true_mask=kwargs[truth_image_name],
                    pred_mask=kwargs['predicted'],
                    title=title.format(index=index, **kwargs))

    def _check_run(self):
        if self.model:
            if self.model._data_reader:
                if self.model.is_compiled:
                    return True
        raise RuntimeError("Cannot run experiment with incomplete model")
        return False


class ExperimentDB(Experiment):  # pragma: no cover

    def __init__(self, dbclient, experiment_id=None, session_id=None,
                 log_base_path='logs',
                 best_model_monitors='val_loss',
                 best_model_modes='auto'):
        """
        An experiment logging performance to a database

        Parameters
        ----------
        dbclient : deoxys.database.DBClient
            The database client
        experiment_id : str, int, or ObjectID depending of the dbclient, optional
            Experiment id, by default None
        session_id : str, int, or ObjectID depending of the dbclient, optional
            Session id, by default None
        log_base_path : str, optional
            Base path to log files, by default 'logs'
        best_model_monitors : str, optional
            Attribute to monitor, by default 'val_loss'
        best_model_modes : str, optional
            One of 'max', 'min', 'auto', by default 'auto'

        Raises
        ------
        ValueError
            When both `session_id` and `experiment_id` are not set
        """
        super().__init__(log_base_path, best_model_monitors, best_model_modes)

        self.dbclient = dbclient

        if experiment_id is None and session_id is None:
            raise ValueError('"session_id" or "experiment_id" must be set')

        if session_id:
            last_model = self.dbclient.find_max(Tables.MODELS, {
                HDF5Attr.SESSION_ID: session_id}, HDF5Attr.EPOCH)
            self.curr_epoch = last_model[HDF5Attr.EPOCH]
            self.session = dbclient.find_by_id(Tables.SESSIONS, session_id)

            self._update_epoch(last_model[HDF5Attr.EPOCH])

            self.from_file(last_model[HDF5Attr.FILE_LOCATION])
        else:
            insert_res = dbclient.insert(Tables.SESSIONS, {
                SessionAttr.EXPERIMENT_ID: experiment_id,
                SessionAttr.CURRENT_EPOCH: 0,
                SessionAttr.STATUS: 'created'
            }, time_logs=True)

            self.session = self.dbclient.find_by_id(
                Tables.SESSIONS, insert_res.inserted_id)

            self.curr_epoch = 0

            experiment_obj = self.dbclient.find_by_id(
                Tables.EXPERIMENTS, experiment_id)

            if ExperimentAttr.CONFIG in experiment_obj:
                self.from_full_config(experiment_obj[ExperimentAttr.CONFIG])
            elif ExperimentAttr.SAVED_MODEL_LOC in experiment_obj:
                self.from_file(experiment_obj[ExperimentAttr.SAVED_MODEL_LOC])

        self.log_base_path = os.path.join(
            log_base_path, str(self.dbclient.get_id(self.session)))

    def run_experiment(self, train_history_log=True,
                       model_checkpoint_period=0,
                       prediction_checkpoint_period=0,
                       save_origin_images=False,
                       verbose=1,
                       epochs=None
                       ):
        log_base_path = self.log_base_path

        if self._check_run():
            if not os.path.exists(log_base_path):
                os.makedirs(log_base_path)

            kwargs = {}

            if epochs:
                kwargs['epochs'] = epochs + self.curr_epoch

            kwargs['initial_epoch'] = self.curr_epoch

            callbacks = []

            if train_history_log:
                callback = self._create_logger(log_base_path,
                                               append=self.curr_epoch > 0)
                callbacks.append(callback)

                callback = self._create_db_logger()
                callbacks.append(callback)

            if model_checkpoint_period > 0:
                if not os.path.exists(log_base_path + self.MODEL_PATH):
                    os.makedirs(log_base_path + self.MODEL_PATH)

                callback = self._create_model_checkpoint(
                    log_base_path,
                    period=model_checkpoint_period)
                callbacks.append(callback)

            if prediction_checkpoint_period > 0:
                if not os.path.exists(log_base_path + self.PREDICTION_PATH):
                    os.makedirs(log_base_path + self.PREDICTION_PATH)

                callback = self._create_prediction_checkpoint(
                    log_base_path,
                    prediction_checkpoint_period,
                    use_original=save_origin_images
                )
                callbacks.append(callback)

            kwargs['callbacks'] = callbacks

            self._update_status(SessionStatus.TRAINING)

            try:
                self.model.fit_train(**kwargs)

                self.curr_epoch += epochs
                self._update_status(SessionStatus.FINISHED)
                self._update_epoch(self.curr_epoch)
            except Exception:
                self._update_status(SessionStatus.FAILED)

            return self

    def _create_db_logger(self):
        return DBLogger(self.dbclient, self.dbclient.get_id(self.session))

    def _update_epoch(self, new_epoch):
        self.dbclient.update_by_id(
            Tables.SESSIONS, self.dbclient.get_id(self.session),
            {SessionAttr.CURRENT_EPOCH: new_epoch}, time_logs=True)

    def _update_status(self, new_status):
        self.dbclient.update_by_id(
            Tables.SESSIONS, self.dbclient.get_id(self.session),
            {SessionAttr.STATUS: new_status}, time_logs=True)

    def _create_model_checkpoint(self, base_path, period):
        return DeoxysModelCheckpoint(
            period=period,
            filepath=base_path + self.MODEL_PATH + self.MODEL_NAME,
            dbclient=self.dbclient,
            session=self.dbclient.get_id(self.session))

    def _create_prediction_checkpoint(self, base_path, period, use_original):
        return PredictionCheckpoint(
            filepath=base_path + self.PREDICTION_PATH + self.PREDICTION_NAME,
            period=period, use_original=use_original,
            dbclient=self.dbclient,
            session=self.dbclient.get_id(self.session))



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import copy
import json
import os
import tensorflow as tf

from .singleton import Singleton
from .file_utils import *
from .json_utils import *
from .data_utils import *


KERAS_STANDALONE = None
TENSORFLOW_EAGER_MODE = None


def deep_copy(obj):
    if type(obj) == dict or type(obj) == list:
        try:
            return json.loads(json.dumps(obj))
        except json.JSONDecodeError:
            pass

    return copy.deepcopy(obj)


def is_keras_standalone():
    global KERAS_STANDALONE
    if KERAS_STANDALONE is None:
        KERAS_STANDALONE = False

        if 'KERAS_MODE' in os.environ:
            mode = os.environ.get('KERAS_MODE')
            if mode.upper() == 'ALONE':
                KERAS_STANDALONE = True
            elif mode.upper() == 'TENSORFLOW':
                KERAS_STANDALONE = False
    return KERAS_STANDALONE


def is_default_tf_eager_mode():
    global TENSORFLOW_EAGER_MODE
    if TENSORFLOW_EAGER_MODE is None:
        TENSORFLOW_EAGER_MODE = (
            not is_keras_standalone()) and tf.__version__.startswith('2.')

    return TENSORFLOW_EAGER_MODE



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest
from deoxys.experiment import Experiment


def test_run_experiment():
    exp = Experiment(
        log_base_path='../../oxford_perf/logs',
        best_model_monitors=['val_loss', 'val_accuracy']
    ).from_full_config(
        'tests/json/sample_dataset_xs_config.json'
    ).run_experiment(
        train_history_log=True,
        model_checkpoint_period=1,
        prediction_checkpoint_period=1,
        epochs=2
    ).plot_performance().plot_prediction(
        masked_images=[i for i in range(10)],
    ).plot_prediction(
        [i for i in range(10)],
        img_name='Prediction_{index:02d}_contour.png',
        contour=True
    ).plot_prediction(
        [i for i in range(10)],
        img_name='Prediction_{index:02d}.png',
        contour=False
    )

    # Best model in this run
    print(exp.best_model())


def test_continue_experiment():
    exp = Experiment(
        log_base_path='../../oxford_perf/logs',
        best_model_monitors=['val_loss', 'val_accuracy']
    ).from_file(
        'tests/h5_files/model.002.h5'
    ).run_experiment(
        train_history_log=True,
        model_checkpoint_period=1,
        prediction_checkpoint_period=1,
        epochs=3,
        initial_epoch=2
    ).plot_performance().plot_prediction(
        masked_images=[i for i in range(10)],
    )


def test_run_test():
    exp = Experiment(
        log_base_path='../../oxford_perf/logs',
        best_model_monitors=['val_loss', 'val_accuracy']
    ).from_file(
        'tests/h5_files/model.002.h5'
    ).run_test(masked_images=[i for i in range(10)])



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest
import h5py
import os
import numpy as np
from deoxys.data import DataReaders, DataReader, HDF5Reader, H5Reader, \
    datareader_from_config, BasePreprocessor
from deoxys.customize import register_datareader, \
    unregister_datareader, custom_datareader, custom_preprocessor
from deoxys.utils import Singleton


VALID_H5_FILE = 'file10.h5'
VALID_H5_FILE_M = 'file10_M.h5'
VALID_H5_FILE_UNIQUE = 'file10_U.h5'
INVALID_H5_FILE_NO_GROUPS = 'file_nogroups.h5'
INVALID_H5_FILE_GROUPS = 'invalid_file.h5'


@pytest.fixture(autouse=True)
def clear_singleton():
    Singleton._instances = {}  # clear singleton


def setup_data():
    # Setup the hdf5 files
    data = np.arange(25000)
    data = np.reshape(data, (1000, 5, 5))
    target = [1, 2, 3, 4, 5] * 200

    return data, target


def teardown(filename=''):
    if os.path.isfile(filename):
        os.remove(filename)


@pytest.fixture(scope='module', autouse=True)
def h5file():
    # Create standardized h5 file with 10 folds, named fold_x
    filename = VALID_H5_FILE
    data, target = setup_data()

    hf = h5py.File(filename, 'w')

    for i in range(10):
        start, end = i * 100, (i + 1) * 100
        group = hf.create_group('fold_{0}'.format(i))
        group.create_dataset('input', data=data[start: end])
        group.create_dataset('target', data=target[start: end])

    hf.close()
    yield filename
    teardown(filename)


@pytest.fixture(scope='module', autouse=True)
def h5file_unique():
    # Create standardized h5 file with 10 folds, named fold_x
    filename = VALID_H5_FILE_UNIQUE
    data = np.arange(25000)
    data = np.reshape(data, (1000, 5, 5))
    target = list(np.arange(1, 1001))

    with h5py.File(filename, 'w') as hf:
        for i in range(10):
            start, end = i * 100, (i + 1) * 100
            group = hf.create_group('fold_{0}'.format(i))
            group.create_dataset('input', data=data[start: end])
            group.create_dataset('target', data=target[start: end])

    yield filename
    teardown(filename)


@pytest.fixture(scope='module', autouse=True)
def h5file_modified():
    # Create standardized h5 file with 10 folds, named randomly
    filename = VALID_H5_FILE_M
    data, target = setup_data()

    hf = h5py.File(filename, 'w')

    for i in range(10):
        start, end = i * 100, (i + 1) * 100
        group = hf.create_group(str(i))
        group.create_dataset('input', data=data[start: end])
        group.create_dataset('target', data=target[start: end])

    hf.close()
    yield filename

    teardown(filename)


@pytest.fixture(scope='module', autouse=True)
def h5file_nogroups():
    # Create invalid h5 file with no groups
    filename = INVALID_H5_FILE_NO_GROUPS
    data, target = setup_data()

    hf = h5py.File(filename, 'w')
    hf.close()
    yield filename

    teardown(filename)


@pytest.fixture(scope='module', autouse=True)
def h5file_invalid():
    # Create invalid h5 file with different datasets in each groups
    filename = INVALID_H5_FILE_GROUPS
    data, target = setup_data()

    hf = h5py.File(filename, 'w')

    # First 3 valid groups
    for i in range(3):
        start, end = i * 100, (i + 1) * 100
        group = hf.create_group('fold_{0}'.format(i))
        group.create_dataset('input', data=data[start: end])
        group.create_dataset('target', data=target[start: end])

    # Next 3 groups with different structures
    for i in range(3, 6):
        start, end = i * 100, (i + 1) * 100
        group = hf.create_group('fold_{0}'.format(i))
        group.create_dataset('x', data=data[start: end])
        group.create_dataset('y', data=target[start: end])

    # Last 4 groups with additional data
    for i in range(6, 10):
        start, end = i * 100, (i + 1) * 100
        group = hf.create_group('fold_{0}'.format(i))
        group.create_dataset('x', data=data[start: end])
        group.create_dataset('y', data=target[start: end])
        group.create_dataset('z', data=target[start: end])

    hf.close()
    yield filename

    teardown(filename)


@pytest.fixture
def datareader_class():
    class TestDataReader(DataReader):
        pass

    yield TestDataReader


def test_is_singleton():
    datareaders_instance = DataReaders()
    another_instance = DataReaders()
    assert datareaders_instance is another_instance


def test_register_random_obj():
    with pytest.raises(ValueError):
        register_datareader("TestDataReader", object)


def test_register_datareader_success(datareader_class):
    register_datareader("TestDataReader", datareader_class)

    assert DataReaders()._dataReaders["TestDataReader"] is datareader_class


def test_register_duplicate_datareader(datareader_class):
    register_datareader("TestDataReader", datareader_class)

    with pytest.raises(KeyError):
        register_datareader("TestDataReader", datareader_class)


def test_unregister_datareader(datareader_class):
    register_datareader("TestDataReader", datareader_class)

    assert DataReaders()._dataReaders["TestDataReader"] is datareader_class
    unregister_datareader("TestDataReader")

    assert "TestDataReader" not in DataReaders()._dataReaders


def test_decorator():
    @custom_datareader
    class TestDataReader2(DataReader):
        pass

    assert DataReaders()._dataReaders["TestDataReader2"] is TestDataReader2


def test_hdf5_dr_constructor_invalid_preprocessor():
    dr = HDF5Reader(VALID_H5_FILE, batch_size=8, preprocessors=[1],
                    x_name='input', y_name='target')

    with pytest.raises(ValueError):
        dr.train_generator

    with pytest.raises(ValueError):
        dr.val_generator

    with pytest.raises(ValueError):
        dr.test_generator


def test_hdf5_dr_constructor():
    dr = HDF5Reader(VALID_H5_FILE, batch_size=8,
                    x_name='input', y_name='target', fold_prefix='fold')
    assert dr.train_folds == ['fold_0']
    assert dr.val_folds == ['fold_1']
    assert dr.test_folds == ['fold_2']
    dr.train_generator
    dr.test_generator
    dr.val_generator


def test_invalid_dr_from_config():
    with pytest.raises(ValueError):
        datareader_from_config({
            'config': {}
        })


def test_hdf5_dr_from_config():
    config = {
        'filename': VALID_H5_FILE,
        'x_name': 'input',
        'y_name': 'target',
        'train_folds': [0, 1, 2],
        'val_folds': [3],
        'test_folds': [4, 5]
    }

    dr = datareader_from_config({
        'class_name': 'HDF5Reader',
        'config': config
    })

    assert dr.train_folds == ['fold_0', 'fold_1', 'fold_2']
    assert dr.val_folds == ['fold_3']
    assert dr.test_folds == ['fold_4', 'fold_5']

    dr.train_generator
    dr.test_generator
    dr.val_generator


def test_hdf5_dr_constructor_no_prefix():
    dr = HDF5Reader(VALID_H5_FILE_M, batch_size=8,
                    x_name='input', y_name='target', fold_prefix=None)
    assert dr.train_folds == [0]
    assert dr.val_folds == [1]
    assert dr.test_folds == [2]
    dr.train_generator
    dr.test_generator
    dr.val_generator


def test_hdf5_dr_constructor_invalid_xname():
    dr = HDF5Reader(VALID_H5_FILE, batch_size=8,
                    x_name='x', y_name='target')

    with pytest.raises(RuntimeError):
        dr.train_generator


def test_hdf5_dr_constructor_invalid_yname():
    with pytest.raises(RuntimeError):
        dr = HDF5Reader(VALID_H5_FILE, batch_size=8,
                        x_name='input', y_name='y')
        dr.train_generator


def test_hdf5_dr_constructor_h5file_no_groups():
    dr = HDF5Reader(INVALID_H5_FILE_NO_GROUPS, batch_size=8,
                    x_name='input', y_name='target')
    with pytest.raises(RuntimeError):
        dr.train_generator


def test_hdf5_dr_constructor_invalid_datasetname():
    # 0 - 2: input, target
    # 3 - 5: x, y
    # 6 - 10: x, y, z
    dr = HDF5Reader(INVALID_H5_FILE_GROUPS, batch_size=8,
                    x_name='input', y_name='target',
                    train_folds=[0, 1, 2], val_folds=[3, 4, 5],
                    test_folds=[6, 7, 8])

    dr.train_generator
    with pytest.raises(RuntimeError):
        dr.val_generator

    with pytest.raises(RuntimeError):
        dr.test_generator


def test_hdf5_dr_constructor_invalid_h5_structure():
    # 0 - 2: input, target
    # 3 - 5: x, y
    # 6 - 10: x, y, z
    dr = HDF5Reader(INVALID_H5_FILE_GROUPS, batch_size=8,
                    x_name='x', y_name='y',
                    train_folds=[3, 4, 5, 6], val_folds=[7],
                    test_folds=[8, 9])

    with pytest.raises(RuntimeError):
        dr.train_generator

    dr.test_generator
    dr.val_generator


def test_hdf5_dr_total_batch():
    dr = HDF5Reader(VALID_H5_FILE, batch_size=8,
                    x_name='input', y_name='target',
                    train_folds=[0, 1, 2, 3], val_folds=[4],
                    test_folds=[5, 6])

    assert dr.train_generator.total_batch == int(np.ceil(100 / 8) * 4)
    assert dr.val_generator.total_batch == int(np.ceil(100 / 8))
    assert dr.test_generator.total_batch == int(np.ceil(100 / 8) * 2)


def test_hdf5_dr_generator():
    batch_size = 8
    dr = HDF5Reader(VALID_H5_FILE, batch_size=batch_size,
                    x_name='input', y_name='target',
                    train_folds=[0, 1, 2, 3], val_folds=[4],
                    test_folds=[5, 6])

    expected_train_input = np.reshape(np.arange(400 * 25), (400, 5, 5))
    expected_train_target = np.array([1, 2, 3, 4, 5] * 80)

    total_batch = dr.train_generator.total_batch
    data_gen = dr.train_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        print(start, end, expected_train_target)
        print(target)

        assert np.all(expected_train_target[start:end] == target)
        assert np.all(expected_train_input[start:end] == input_data)


def test_hdf5_dr_generator_preprocessor():
    batch_size = 8

    @custom_preprocessor
    class PlusOnePreprocessor(BasePreprocessor):
        def transform(self, x, y):
            return x + 1, y + 1

    dr = HDF5Reader(VALID_H5_FILE, batch_size=batch_size,
                    x_name='input', y_name='target',
                    train_folds=[0, 1, 2, 3], val_folds=[4],
                    test_folds=[5, 6],
                    preprocessors=PlusOnePreprocessor())

    expected_train_input = np.reshape(np.arange(1, 400 * 25 + 1), (400, 5, 5))
    expected_train_target = np.array([2, 3, 4, 5, 6] * 80)

    total_batch = dr.train_generator.total_batch
    data_gen = dr.train_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        print(start, end, expected_train_target)
        print(target)

        assert np.all(expected_train_target[start:end] == target)
        assert np.all(expected_train_input[start:end] == input_data)


def test_hdf5_dr_generator_multipreprocessors():
    batch_size = 8

    @custom_preprocessor
    class PlusOnePreprocessor(BasePreprocessor):
        def transform(x, y):
            return x + 1, y + 1

    dr = HDF5Reader(VALID_H5_FILE, batch_size=batch_size,
                    x_name='input', y_name='target',
                    train_folds=[0, 1, 2, 3], val_folds=[4],
                    test_folds=[5, 6],
                    preprocessors=[PlusOnePreprocessor, PlusOnePreprocessor])

    expected_train_input = np.reshape(np.arange(2, 400 * 25 + 2), (400, 5, 5))
    expected_train_target = np.array([3, 4, 5, 6, 7] * 80)

    total_batch = dr.train_generator.total_batch
    data_gen = dr.train_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        print(start, end, expected_train_target)
        print(target)

        assert np.all(expected_train_target[start:end] == target)
        assert np.all(expected_train_input[start:end] == input_data)


def test_hdf5_dr_original_test():
    batch_size = 8
    dr = HDF5Reader(VALID_H5_FILE, batch_size=batch_size,
                    x_name='input', y_name='target',
                    train_folds=[0, 1, 2, 3], val_folds=[4],
                    test_folds=[5, 6])
    original_test = dr.original_test
    assert len(original_test.keys()) == 2
    assert 'input' in original_test
    assert original_test['input'].shape[0] == 200
    assert 'target' in original_test
    assert original_test['target'].shape[0] == 200


def test_hdf5_dr_original_test_more_column():
    batch_size = 8
    dr = HDF5Reader(INVALID_H5_FILE_GROUPS, batch_size=batch_size,
                    x_name='x', y_name='y',
                    train_folds=[6, 7], val_folds=[8],
                    test_folds=[9])
    original_test = dr.original_test
    assert len(original_test.keys()) == 3
    assert 'x' in original_test
    assert original_test['x'].shape[0] == 100
    assert 'y' in original_test
    assert original_test['y'].shape[0] == 100
    assert 'z' in original_test
    assert original_test['z'].shape[0] == 100


def test_hdf5_dr_original_val():
    batch_size = 8
    dr = HDF5Reader(VALID_H5_FILE, batch_size=batch_size,
                    x_name='input', y_name='target',
                    train_folds=[0, 1, 2, 3], val_folds=[4, 5],
                    test_folds=[6, 7, 8])
    original_val = dr.original_val
    assert len(original_val.keys()) == 2
    assert 'input' in original_val
    assert original_val['input'].shape[0] == 200
    assert 'target' in original_val
    assert original_val['target'].shape[0] == 200


def test_hdf5_dr_original_val_more_column():
    batch_size = 8
    dr = HDF5Reader(INVALID_H5_FILE_GROUPS, batch_size=batch_size,
                    x_name='x', y_name='y',
                    train_folds=[6, 7], val_folds=[8],
                    test_folds=[9])
    original_val = dr.original_val
    assert len(original_val.keys()) == 3
    assert 'x' in original_val
    assert original_val['x'].shape[0] == 100
    assert 'y' in original_val
    assert original_val['y'].shape[0] == 100
    assert 'z' in original_val
    assert original_val['z'].shape[0] == 100


def test_h5_dr_constructor_invalid_preprocessor():
    dr = H5Reader(VALID_H5_FILE, batch_size=8, preprocessors=[1],
                  x_name='input', y_name='target')

    with pytest.raises(ValueError):
        dr.train_generator

    with pytest.raises(ValueError):
        dr.val_generator

    with pytest.raises(ValueError):
        dr.test_generator


def test_h5_dr_constructor():
    dr = H5Reader(VALID_H5_FILE, batch_size=8,
                  x_name='input', y_name='target', fold_prefix='fold')
    assert dr.train_folds == ['fold_0']
    assert dr.val_folds == ['fold_1']
    assert dr.test_folds == ['fold_2']
    dr.train_generator
    dr.test_generator
    dr.val_generator


def test_h5_dr_from_config():
    config = {
        'filename': VALID_H5_FILE,
        'x_name': 'input',
        'y_name': 'target',
        'train_folds': [0, 1, 2],
        'val_folds': [3],
        'test_folds': [4, 5]
    }

    dr = datareader_from_config({
        'class_name': 'H5Reader',
        'config': config
    })

    assert dr.train_folds == ['fold_0', 'fold_1', 'fold_2']
    assert dr.val_folds == ['fold_3']
    assert dr.test_folds == ['fold_4', 'fold_5']

    dr.train_generator
    dr.test_generator
    dr.val_generator


def test_h5_dr_constructor_no_prefix():
    dr = H5Reader(VALID_H5_FILE_M, batch_size=8,
                  x_name='input', y_name='target', fold_prefix=None)
    assert dr.train_folds == [0]
    assert dr.val_folds == [1]
    assert dr.test_folds == [2]
    dr.train_generator
    dr.test_generator
    dr.val_generator


def test_h5_dr_constructor_invalid_xname():
    dr = H5Reader(VALID_H5_FILE, batch_size=8,
                  x_name='x', y_name='target')

    with pytest.raises(RuntimeError):
        dr.train_generator


def test_h5_dr_constructor_invalid_yname():
    with pytest.raises(RuntimeError):
        dr = H5Reader(VALID_H5_FILE, batch_size=8,
                      x_name='input', y_name='y')
        dr.train_generator


def test_h5_dr_constructor_h5file_no_groups():
    dr = H5Reader(INVALID_H5_FILE_NO_GROUPS, batch_size=8,
                  x_name='input', y_name='target')
    with pytest.raises(RuntimeError):
        dr.train_generator


def test_h5_dr_constructor_invalid_datasetname():
    # 0 - 2: input, target
    # 3 - 5: x, y
    # 6 - 10: x, y, z
    dr = H5Reader(INVALID_H5_FILE_GROUPS, batch_size=8,
                  x_name='input', y_name='target',
                  train_folds=[0, 1, 2], val_folds=[3, 4, 5],
                  test_folds=[6, 7, 8])

    dr.train_generator
    with pytest.raises(RuntimeError):
        dr.val_generator

    with pytest.raises(RuntimeError):
        dr.test_generator


def test_h5_dr_constructor_invalid_h5_structure():
    # 0 - 2: input, target
    # 3 - 5: x, y
    # 6 - 10: x, y, z
    dr = H5Reader(INVALID_H5_FILE_GROUPS, batch_size=8,
                  x_name='x', y_name='y',
                  train_folds=[3, 4, 5, 6], val_folds=[7],
                  test_folds=[8, 9])

    with pytest.raises(RuntimeError):
        dr.train_generator

    dr.test_generator
    dr.val_generator


def test_h5_dr_total_batch():
    dr = H5Reader(VALID_H5_FILE, batch_size=8,
                  x_name='input', y_name='target',
                  train_folds=[0, 1, 2, 3], val_folds=[4],
                  test_folds=[5, 6])

    assert dr.train_generator.total_batch == int(np.ceil(100 / 8) * 4)
    assert dr.val_generator.total_batch == int(np.ceil(100 / 8))
    assert dr.test_generator.total_batch == int(np.ceil(100 / 8) * 2)


def test_h5_dr_generator():
    batch_size = 8
    dr = H5Reader(VALID_H5_FILE, batch_size=batch_size,
                  x_name='input', y_name='target',
                  train_folds=[0, 1, 2, 3], val_folds=[4],
                  test_folds=[5, 6])

    expected_train_input = np.reshape(np.arange(400 * 25), (400, 5, 5))
    expected_train_target = np.array([1, 2, 3, 4, 5] * 80)

    total_batch = dr.train_generator.total_batch
    data_gen = dr.train_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        print(start, end, expected_train_target)
        print(target)

        assert np.all(expected_train_target[start:end] == target)
        assert np.all(expected_train_input[start:end] == input_data)


def test_h5_dr_generator_preprocessor():
    batch_size = 8

    @custom_preprocessor
    class PlusOnePreprocessor(BasePreprocessor):
        def transform(self, x, y):
            return x + 1, y + 1

    dr = H5Reader(VALID_H5_FILE, batch_size=batch_size,
                  x_name='input', y_name='target',
                  train_folds=[0, 1, 2, 3], val_folds=[4],
                  test_folds=[5, 6],
                  preprocessors=PlusOnePreprocessor())

    expected_train_input = np.reshape(np.arange(1, 400 * 25 + 1), (400, 5, 5))
    expected_train_target = np.array([2, 3, 4, 5, 6] * 80)

    total_batch = dr.train_generator.total_batch
    data_gen = dr.train_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        print(start, end, expected_train_target)
        print(target)

        assert np.all(expected_train_target[start:end] == target)
        assert np.all(expected_train_input[start:end] == input_data)


def test_h5_dr_generator_multipreprocessors():
    batch_size = 8

    @custom_preprocessor
    class PlusOnePreprocessor(BasePreprocessor):
        def transform(x, y):
            return x + 1, y + 1

    dr = H5Reader(VALID_H5_FILE, batch_size=batch_size,
                  x_name='input', y_name='target',
                  train_folds=[0, 1, 2, 3], val_folds=[4],
                  test_folds=[5, 6],
                  preprocessors=[PlusOnePreprocessor, PlusOnePreprocessor])

    expected_train_input = np.reshape(np.arange(2, 400 * 25 + 2), (400, 5, 5))
    expected_train_target = np.array([3, 4, 5, 6, 7] * 80)

    total_batch = dr.train_generator.total_batch
    data_gen = dr.train_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        print(start, end, expected_train_target)
        print(target)

        assert np.all(expected_train_target[start:end] == target)
        assert np.all(expected_train_input[start:end] == input_data)


def test_h5_dr_generator_augmentation():
    batch_size = 8

    @custom_preprocessor
    class PlusOnePreprocessor(BasePreprocessor):
        def transform(self, x, y):
            return x + 1, y + 1

    dr = H5Reader(VALID_H5_FILE, batch_size=batch_size,
                  x_name='input', y_name='target',
                  train_folds=[0, 1, 2, 3], val_folds=[4],
                  test_folds=[5, 6],
                  preprocessors=PlusOnePreprocessor(),
                  augmentations=PlusOnePreprocessor())

    expected_train_input = np.reshape(np.arange(2, 400 * 25 + 2), (400, 5, 5))
    expected_train_target = np.array([3, 4, 5, 6, 7] * 80)

    total_batch = dr.train_generator.total_batch
    data_gen = dr.train_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        print(start, end, expected_train_target)
        print(target)

        assert np.all(expected_train_target[start:end] == target)
        assert np.all(expected_train_input[start:end] == input_data)

    expected_val_input = np.reshape(
        np.arange(400 * 25 + 1, 500 * 25 + 1), (100, 5, 5))
    expected_val_target = np.array([2, 3, 4, 5, 6] * 20)

    total_batch = dr.val_generator.total_batch
    data_gen = dr.val_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        print(start, end, expected_val_target)
        print(target)

        assert np.all(expected_val_target[start:end] == target)
        assert np.all(expected_val_input[start:end] == input_data)


def test_h5_dr_generator_multiaugmentations():
    batch_size = 8

    @custom_preprocessor
    class PlusOnePreprocessor(BasePreprocessor):
        def transform(x, y):
            return x + 1, y + 1

    dr = H5Reader(VALID_H5_FILE, batch_size=batch_size,
                  x_name='input', y_name='target',
                  train_folds=[0, 1, 2, 3], val_folds=[4],
                  test_folds=[5, 6],
                  preprocessors=[PlusOnePreprocessor, PlusOnePreprocessor],
                  augmentations=[PlusOnePreprocessor, PlusOnePreprocessor])

    expected_train_input = np.reshape(np.arange(4, 400 * 25 + 4), (400, 5, 5))
    expected_train_target = np.array([5, 6, 7, 8, 9] * 80)

    total_batch = dr.train_generator.total_batch
    data_gen = dr.train_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        print(start, end, expected_train_target)
        print(target)

        assert np.all(expected_train_target[start:end] == target)
        assert np.all(expected_train_input[start:end] == input_data)

    expected_val_input = np.reshape(
        np.arange(400 * 25 + 2, 500 * 25 + 2), (100, 5, 5))
    expected_val_target = np.array([3, 4, 5, 6, 7] * 20)

    total_batch = dr.val_generator.total_batch
    data_gen = dr.val_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        print(start, end, expected_val_target)
        print(target)

        assert np.all(expected_val_target[start:end] == target)
        assert np.all(expected_val_input[start:end] == input_data)


def test_h5_dr_original_test():
    batch_size = 8
    dr = H5Reader(VALID_H5_FILE, batch_size=batch_size,
                  x_name='input', y_name='target',
                  train_folds=[0, 1, 2, 3], val_folds=[4],
                  test_folds=[5, 6])
    original_test = dr.original_test
    assert len(original_test.keys()) == 2
    assert 'input' in original_test
    assert original_test['input'].shape[0] == 200
    assert 'target' in original_test
    assert original_test['target'].shape[0] == 200


def test_h5_dr_original_test_more_column():
    batch_size = 8
    dr = H5Reader(INVALID_H5_FILE_GROUPS, batch_size=batch_size,
                  x_name='x', y_name='y',
                  train_folds=[6, 7], val_folds=[8],
                  test_folds=[9])
    original_test = dr.original_test
    assert len(original_test.keys()) == 3
    assert 'x' in original_test
    assert original_test['x'].shape[0] == 100
    assert 'y' in original_test
    assert original_test['y'].shape[0] == 100
    assert 'z' in original_test
    assert original_test['z'].shape[0] == 100


def test_h5_dr_original_val():
    batch_size = 8
    dr = H5Reader(VALID_H5_FILE, batch_size=batch_size,
                  x_name='input', y_name='target',
                  train_folds=[0, 1, 2, 3], val_folds=[4, 5],
                  test_folds=[6, 7, 8])
    original_val = dr.original_val
    assert len(original_val.keys()) == 2
    assert 'input' in original_val
    assert original_val['input'].shape[0] == 200
    assert 'target' in original_val
    assert original_val['target'].shape[0] == 200


def test_h5_dr_original_val_more_column():
    batch_size = 8
    dr = H5Reader(INVALID_H5_FILE_GROUPS, batch_size=batch_size,
                  x_name='x', y_name='y',
                  train_folds=[6, 7], val_folds=[8],
                  test_folds=[9])
    original_val = dr.original_val
    assert len(original_val.keys()) == 3
    assert 'x' in original_val
    assert original_val['x'].shape[0] == 100
    assert 'y' in original_val
    assert original_val['y'].shape[0] == 100
    assert 'z' in original_val
    assert original_val['z'].shape[0] == 100


def test_h5_dr_shuffle():
    batch_size = 8

    original_input = np.arange(25000)
    original_input = np.reshape(original_input, (1000, 5, 5))

    dr = H5Reader(VALID_H5_FILE_UNIQUE, batch_size=batch_size,
                  x_name='input', y_name='target',
                  train_folds=[0, 1, 2, 3], val_folds=[4],
                  test_folds=[5, 6], shuffle=True)

    total_batch = dr.train_generator.total_batch
    data_gen = dr.train_generator.generate()

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        for x, y in zip(input_data, target):
            # The target is in range(1-1000), thus can act as the index
            assert np.all(x == original_input[y-1])

    # No shuffling in val and test data
    expected_val_input = np.reshape(
        np.arange(400 * 25, 500 * 25), (100, 5, 5))
    expected_val_target = np.array(np.arange(401, 501))

    total_batch = dr.val_generator.total_batch
    data_gen = dr.val_generator.generate()
    end = 0

    for i, (input_data, target) in enumerate(data_gen):
        if i >= total_batch:
            break

        start = end
        end = start + batch_size

        if end % 100 < batch_size:
            end -= end % 100

        assert np.all(expected_val_target[start:end] == target)
        assert np.all(expected_val_input[start:end] == input_data)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import numpy as np

from ..keras import backend as K
from ..keras.losses import Loss, deserialize
from ..utils import Singleton, is_keras_standalone


class BinaryFbetaLoss(Loss):
    def __init__(self, reduction='auto', name="binary_fbeta", beta=1):
        if is_keras_standalone():
            # use Keras default reduction
            super().__init__('sum_over_batch_size', name)
        else:
            super().__init__(reduction, name)

        self.beta = beta

    def call(self, target, prediction):
        size = len(prediction.get_shape().as_list())
        reduce_ax = list(range(1, size))
        eps = 1e-8

        true_positive = K.sum(prediction * target, axis=reduce_ax)
        target_positive = K.sum(K.square(target), axis=reduce_ax)
        predicted_positive = K.sum(
            K.square(prediction), axis=reduce_ax)

        fb_numerator = (1 + self.beta ** 2) * true_positive + eps
        fb_denominator = (
            (self.beta ** 2) * target_positive + predicted_positive + eps
        )

        return 1 - fb_numerator / fb_denominator


class ModifiedDiceLoss(Loss):
    def __init__(self, reduction='auto', name="modified_dice_loss", beta=1):
        if is_keras_standalone():
            # use Keras default reduction
            super().__init__('sum_over_batch_size', name)
        else:
            super().__init__(reduction, name)

        self.beta = beta

    def call(self, target, prediction):
        size = len(prediction.get_shape().as_list())
        reduce_ax = list(range(1, size))
        eps = 1e-8

        true_positive = K.sum(prediction * target, axis=reduce_ax)
        target_positive = K.sum(target, axis=reduce_ax)
        predicted_positive = K.sum(
            prediction, axis=reduce_ax)

        fb_numerator = (1 + self.beta ** 2) * true_positive + eps
        fb_denominator = (
            (self.beta ** 2) * target_positive + predicted_positive + eps
        )

        return 1 - fb_numerator / fb_denominator


class Losses(metaclass=Singleton):
    """
    A singleton that contains all the registered customized losses
    """

    def __init__(self):
        self._losses = {
            'BinaryFbetaLoss': BinaryFbetaLoss,
            'ModifiedDiceLoss': ModifiedDiceLoss
        }

    def register(self, key, loss):
        if not issubclass(loss, Loss):
            raise ValueError(
                "The customized loss has to be a subclass"
                + " of keras.losses.Loss"
            )

        if key in self._losses:
            raise KeyError(
                "Duplicated key, please use another key for this loss"
            )
        else:
            self._losses[key] = loss

    def unregister(self, key):
        if key in self._losses:
            del self._losses[key]

    @property
    def losses(self):
        return self._losses


def register_loss(key, loss):
    """
    Register the customized loss.
    If the key name is already registered, it will raise a KeyError exception

    Parameters
    ----------
    key : str
        The unique key-name of the loss
    loss : tensorflow.keras.losses.Loss
        The customized loss class
    """
    Losses().register(key, loss)


def unregister_loss(key):
    """
    Remove the registered loss with the key-name

    Parameters
    ----------
    key : str
        The key-name of the loss to be removed
    """
    Losses().unregister(key)


def loss_from_config(config):
    if type(config) == dict:
        if 'class_name' not in config:
            raise ValueError('class_name is needed to define loss')

        if 'config' not in config:
            # auto add empty config for loss with only class_name
            config['config'] = {}
        return deserialize(
            config,
            custom_objects=Losses().losses)
    return deserialize(config, custom_objects=Losses().losses)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest
import os
import numpy as np
from deoxys.keras.models import Model
from deoxys.keras.layers import Input, Dense, Flatten, Dropout, \
    BatchNormalization, Conv2D, Conv2DTranspose, MaxPooling2D, concatenate, \
    Conv3D, Conv3DTranspose, MaxPooling3D
from deoxys.loaders.architecture import load_architecture
from deoxys.automation import generate_unet_architecture, \
    generate_vnet_architecture, generate_unet_architecture_json, \
    generate_vnet_architecture_json
from deoxys.utils import read_file, load_json_config


FILE_NAME = 'test_architecture.json'


@pytest.fixture(autouse=True)
def clear_files():
    yield
    if os.path.isfile(FILE_NAME):
        os.remove(FILE_NAME)


def check_same_models(actual_model, expected_model):
    actual_config = actual_model.get_config()
    expected_config = expected_model.get_config()
    name_map = [(actual_model.name, expected_model.name)]
    for (actual_layer, expected_layer) in zip(
            actual_model.layers, expected_model.layers):
        name_map.append((actual_layer.name, expected_layer.name))

    actual_config_str = str(actual_config)

    for actual_name, expected_name in name_map:
        actual_config_str = actual_config_str.replace(
            "'" + actual_name + "'", "'" + expected_name + "'")
    print(actual_config_str)
    print('====================')
    print(expected_config)
    return actual_config_str == str(expected_config)


def test_generate_unet_model():
    architecture = generate_unet_architecture(
        n_upsampling=5, n_filter=4, stride=2)

    input_params = {'shape': [128, 128, 3]}

    model = load_architecture(architecture, input_params)
    model.summary()

    def conv_layers(filters, pre_layer):
        conv = BatchNormalization()(
            Conv2D(filters,
                   kernel_size=3,
                   activation='relu',
                   padding='same',
                   kernel_initializer='he_normal')(pre_layer))

        return BatchNormalization()(
            Conv2D(filters,
                   kernel_size=3,
                   activation='relu',
                   padding='same',
                   kernel_initializer='he_normal')(conv))

    input_layer = Input(shape=[128, 128, 3])

    conv_1 = conv_layers(4, input_layer)
    max_pool_1 = MaxPooling2D()(conv_1)

    conv_2 = conv_layers(8, max_pool_1)
    max_pool_2 = MaxPooling2D()(conv_2)

    conv_3 = conv_layers(16, max_pool_2)
    max_pool_3 = MaxPooling2D()(conv_3)

    conv_4 = conv_layers(32, max_pool_3)
    max_pool_4 = MaxPooling2D()(conv_4)

    conv_5 = conv_layers(64, max_pool_4)
    max_pool_5 = MaxPooling2D()(conv_5)

    conv_t_kwargs = {"kernel_size": 3,
                     "strides": 2,
                     "padding": "same",
                     "kernel_initializer": "he_normal"}

    conv_6 = conv_layers(128, max_pool_5)
    conv_trans_1 = Conv2DTranspose(64, **conv_t_kwargs)(conv_6)

    upconv_1 = conv_layers(64, concatenate([conv_5, conv_trans_1]))
    conv_trans_2 = Conv2DTranspose(32, **conv_t_kwargs)(upconv_1)

    upconv_2 = conv_layers(32, concatenate([conv_4, conv_trans_2]))
    conv_trans_3 = Conv2DTranspose(16, **conv_t_kwargs)(upconv_2)

    upconv_3 = conv_layers(16, concatenate([conv_3, conv_trans_3]))
    conv_trans_4 = Conv2DTranspose(8, **conv_t_kwargs)(upconv_3)

    upconv_4 = conv_layers(8, concatenate([conv_2, conv_trans_4]))
    conv_trans_5 = Conv2DTranspose(4, **conv_t_kwargs)(upconv_4)

    upconv_5 = conv_layers(4, concatenate([conv_1, conv_trans_5]))
    output = Conv2D(1, kernel_size=3, activation='sigmoid', padding="same",
                    kernel_initializer="he_normal")(upconv_5)

    expected_model = Model(inputs=input_layer, outputs=output)
    expected_model.summary()

    assert check_same_models(model, expected_model)


def test_generate_unet_model_resize():
    architecture = generate_unet_architecture(
        n_upsampling=5, n_filter=4, stride=2)

    input_params = {'shape': [129, 129, 3]}

    model = load_architecture(architecture, input_params)

    assert np.all(model.input_shape == (None, 129, 129, 3))
    assert np.all(model.output_shape == (None, 129, 129, 1))


def test_create_unet_json():
    unet = generate_unet_architecture()
    generate_unet_architecture_json(FILE_NAME)
    actual = load_json_config(FILE_NAME)

    assert np.all(unet == actual)


def test_generate_vnet_model():
    architecture = generate_vnet_architecture(
        n_upsampling=5, n_filter=4, stride=2)

    input_params = {'shape': [128, 128, 128, 3]}

    model = load_architecture(architecture, input_params)

    def conv_layers(filters, pre_layer):
        conv = BatchNormalization()(
            Conv3D(filters,
                   kernel_size=3,
                   activation='relu',
                   padding='same',
                   kernel_initializer='he_normal')(pre_layer))

        return BatchNormalization()(
            Conv3D(filters,
                   kernel_size=3,
                   activation='relu',
                   padding='same',
                   kernel_initializer='he_normal')(conv))

    input_layer = Input(shape=[128, 128, 128, 3])

    conv_1 = conv_layers(4, input_layer)
    max_pool_1 = MaxPooling3D()(conv_1)

    conv_2 = conv_layers(8, max_pool_1)
    max_pool_2 = MaxPooling3D()(conv_2)

    conv_3 = conv_layers(16, max_pool_2)
    max_pool_3 = MaxPooling3D()(conv_3)

    conv_4 = conv_layers(32, max_pool_3)
    max_pool_4 = MaxPooling3D()(conv_4)

    conv_5 = conv_layers(64, max_pool_4)
    max_pool_5 = MaxPooling3D()(conv_5)

    conv_t_kwargs = {"kernel_size": 3,
                     "strides": 2,
                     "padding": "same",
                     "kernel_initializer": "he_normal"}

    conv_6 = conv_layers(128, max_pool_5)
    conv_trans_1 = Conv3DTranspose(64, **conv_t_kwargs)(conv_6)

    upconv_1 = conv_layers(64, concatenate([conv_5, conv_trans_1]))
    conv_trans_2 = Conv3DTranspose(32, **conv_t_kwargs)(upconv_1)

    upconv_2 = conv_layers(32, concatenate([conv_4, conv_trans_2]))
    conv_trans_3 = Conv3DTranspose(16, **conv_t_kwargs)(upconv_2)

    upconv_3 = conv_layers(16, concatenate([conv_3, conv_trans_3]))
    conv_trans_4 = Conv3DTranspose(8, **conv_t_kwargs)(upconv_3)

    upconv_4 = conv_layers(8, concatenate([conv_2, conv_trans_4]))
    conv_trans_5 = Conv3DTranspose(4, **conv_t_kwargs)(upconv_4)

    upconv_5 = conv_layers(4, concatenate([conv_1, conv_trans_5]))
    output = Conv3D(1, kernel_size=3, activation='sigmoid', padding="same",
                    kernel_initializer="he_normal")(upconv_5)

    expected_model = Model(inputs=input_layer, outputs=output)

    assert check_same_models(model, expected_model)


def test_generate_vnet_model_resize():
    architecture = generate_vnet_architecture(
        n_upsampling=5, n_filter=4, stride=2)

    input_params = {'shape': [129, 129, 129, 3]}

    model = load_architecture(architecture, input_params)

    assert np.all(model.input_shape == (None, 129, 129, 129, 3))
    assert np.all(model.output_shape == (None, 129, 129, 129, 1))



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


from .architecture import load_architecture
from .params import load_params
from .data_loader import load_data
from .train_params import load_train_params




# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import json


def downsampling_block(idx=0, conv_layer='Conv2D', n_filter=64,
                       batchnorm=True, activation='relu',
                       dropout_rate=0,
                       kernel=3, maxpool='MaxPooling2D'):
    block = []

    conv_block = {
        'class_name': conv_layer,
        'config': {
            'filters': n_filter,
            'kernel_size': kernel,
            'activation': activation,
            'kernel_initializer': 'he_normal',
            'padding': 'same'
        }
    }

    if batchnorm:
        conv_block['normalizer'] = {
            'class_name': 'BatchNormalization'
        }

    conv_block_2 = conv_block.copy()
    conv_block_2['name'] = f'conv{idx}'

    block.append(conv_block)
    block.append(conv_block_2)
    block.append({
        'class_name': maxpool
    })

    if dropout_rate > 0:
        block.append({
            'class_name': 'Dropout',
            'config': {
                'rate': dropout_rate
            }
        })

    return block


def bottle_neck(conv_layer='Conv2D', n_filter=64,
                batchnorm=True, activation='relu', kernel=3):
    block = []

    conv_block = {
        'class_name': conv_layer,
        'config': {
            'filters': n_filter,
            'kernel_size': kernel,
            'activation': activation,
            'kernel_initializer': 'he_normal',
            'padding': 'same'
        }
    }

    if batchnorm:
        conv_block['normalizer'] = {
            'class_name': 'BatchNormalization'
        }

    block.append(conv_block)
    block.append(conv_block)

    return block


def upsampling_block(idx=0, conv_layer='Conv2D', n_filter=64,
                     batchnorm=True, activation='relu',
                     dropout_rate=0,
                     kernel=3, stride=1):
    block = []
    upconv_block = {
        'name': f'upconv{idx}',
        'class_name': f'{conv_layer}Transpose',
        'config': {
            'filters': n_filter,
            'kernel_size': kernel,
            'strides': stride,
            'kernel_initializer': 'he_normal',
            'padding': 'same'
        }
    }

    block.append(upconv_block)

    if dropout_rate > 0:
        block.append({
            'class_name': 'Dropout',
            'config': {
                'rate': dropout_rate
            },
            'inputs': [
                f'conv{idx}',
                f'upconv{idx}'
            ]
        })

    conv_block = {
        'class_name': conv_layer,
        'config': {
            'filters': n_filter,
            'kernel_size': kernel,
            'activation': activation,
            'kernel_initializer': 'he_normal',
            'padding': 'same'
        }
    }

    if batchnorm:
        conv_block['normalizer'] = {
            'class_name': 'BatchNormalization'
        }
    conv_block_2 = conv_block.copy()

    if dropout_rate == 0:
        conv_block['inputs'] = [
            f'conv{idx}',
            f'upconv{idx}'
        ]

    block.append(conv_block)
    block.append(conv_block_2)

    return block


def _generate_unet_layers(postfix, n_upsampling=4, n_filter=64,
                          batchnorm=True, activation='relu',
                          dropout_rate=0,
                          kernel=3, stride=1, n_class=2):
    conv_layer = f'Conv{postfix}'
    maxpool = f'MaxPooling{postfix}'
    layers = []

    if '__iter__' not in dir(n_filter):
        n_filter = [n_filter * (2**i) for i in range(n_upsampling + 1)]

    for i in range(n_upsampling):
        layers.extend(downsampling_block(
            idx=i + 1, conv_layer=conv_layer, n_filter=n_filter[i],
            batchnorm=batchnorm,
            activation=activation, dropout_rate=dropout_rate, kernel=kernel,
            maxpool=maxpool))

    layers.extend(bottle_neck(conv_layer=conv_layer,
                              n_filter=n_filter[-1],
                              batchnorm=batchnorm,
                              activation=activation,
                              kernel=kernel))

    for i in range(n_upsampling, 0, -1):
        layers.extend(upsampling_block(idx=i, conv_layer=conv_layer,
                                       n_filter=n_filter[i - 1],
                                       batchnorm=batchnorm,
                                       activation=activation,
                                       dropout_rate=dropout_rate,
                                       kernel=kernel, stride=stride))

    if n_class <= 2:
        layers.append({
            'class_name': conv_layer,
            'config': {
                'filters': 1,
                'kernel_size': kernel,
                'activation': 'sigmoid',
                'kernel_initializer': 'he_normal',
                'padding': 'same'
            }
        })
    else:
        layers.append({
            'class_name': conv_layer,
            'config': {
                'filters': n_class,
                'kernel_size': kernel,
                'activation': 'softmax',
                'kernel_initializer': 'he_normal',
                'padding': 'same'
            }
        })

    return layers


def generate_unet_architecture(n_upsampling=4, n_filter=64,
                               batchnorm=True, activation='relu',
                               dropout_rate=0,
                               kernel=3, stride=1, n_class=2):

    return {
        'type': 'Unet',
        'layers': _generate_unet_layers('2D', n_upsampling, n_filter,
                                        batchnorm, activation,
                                        dropout_rate,
                                        kernel, stride, n_class)
    }


def generate_vnet_architecture(n_upsampling=4, n_filter=64,
                               batchnorm=True, activation='relu',
                               dropout_rate=0,
                               kernel=3, stride=1, n_class=2):

    return {
        'type': 'Vnet',
        'layers': _generate_unet_layers('3D', n_upsampling, n_filter,
                                        batchnorm, activation,
                                        dropout_rate,
                                        kernel, stride, n_class)
    }


def generate_unet_architecture_json(filename, n_upsampling=4, n_filter=64,
                                    batchnorm=True, activation='relu',
                                    dropout_rate=0,
                                    kernel=3, stride=1, n_class=2):

    with open(filename, 'w') as f:
        json.dump(generate_unet_architecture(n_upsampling, n_filter,
                                             batchnorm, activation,
                                             dropout_rate,
                                             kernel, stride, n_class), f)


def generate_vnet_architecture_json(filename, n_upsampling=4, n_filter=64,
                                    batchnorm=True, activation='relu',
                                    dropout_rate=0,
                                    kernel=3, stride=1, n_class=2):

    with open(filename, 'w') as f:
        json.dump(generate_vnet_architecture(n_upsampling, n_filter,
                                             batchnorm, activation,
                                             dropout_rate,
                                             kernel, stride, n_class), f)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import h5py
import numpy as np
from deoxys.keras.preprocessing import ImageDataGenerator
from .data_generator import DataGenerator, HDF5DataGenerator, H5DataGenerator
from ..utils import Singleton, file_finder


class DataReader:
    """
    The base class of the Data Reader. Any newly created DataReader will
    inherit from this class.
    """

    def __init__(self, *args, **kwargs):
        # the existence of the data reader is True by default
        # if the data reader cannot be loaded because of IO reason,
        # set this value to false
        self.ready = True

    @property
    def train_generator(self):
        """
        Data Generator for the training dataset

        Returns
        -------
        deoxys.data.DataGenerator
            An DataGenerator instance that generates the train dataset
        """
        return DataGenerator().generate()

    @property
    def test_generator(self):
        """
        Data Generator for the test dataset

        Returns
        -------
        deoxys.data.DataGenerator
            An DataGenerator instance that generates the test dataset
        """
        return DataGenerator().generate()

    @property
    def val_generator(self):
        """
        Data Generator for the validation dataset

        Returns
        -------
        deoxys.data.DataGenerator
            An DataGenerator instance that generates the validation dataset
        """
        return DataGenerator().generate()

    @property
    def original_test(self):
        pass


class HDF5Reader(DataReader):
    """DataReader that use data from an hdf5 file.

        Initialize a HDF5 Data Reader, which reads data from a HDF5
        file. This file should be split into groups. Each group contain
        datasets, each of which is a column in the data.

        Example:

        The dataset X contain 1000 samples, with 4 columns:
        x, y, z, t. Where x is the main input, y and z are supporting
        information (index, descriptions) and t is the target for
        prediction. We want to test 30% of this dataset, and have a
        cross validation of 100 samples.

        Then, the hdf5 containing dataset X should have 10 groups,
        each group contains 100 samples. We can name these groups
        'fold_1', 'fold_2', 'fold_3', ... , 'fold_9', 'fold_10'.
        Each group will then have 4 datasets: x, y, z and t, each of
        which has 100 items.

        Since x is the main input, then `x_name='x'`, and t is the
        target for prediction, then `y_name='t'`. We named the groups
        in the form of fold_n, then `fold_prefix='fold'`.

        Let's assume the data is stratified, we want to test on the
        last 30% of the data, so `test_folds=[8, 9, 10]`.
        100 samples is used for cross-validation. Thus, one option for
        `train_folds` and `val_folds` is `train_folds=[1,2,3,4,5,6]`
        and `val_folds=[7]`. Or in another experiment, you can set
        `train_folds=[2,3,4,5,6,7]` and `val_folds=[1]`.

        If the hdf5 didn't has any formular for group name, then you
        can set `fold_prefix=None` then put the full group name
        directly to `train_folds`, `val_folds` and `test_folds`.

        Parameters
        ----------
        filename : str
            The hdf5 file name that contains the data.
        batch_size : int, optional
            Number of sample to feeds in
            the neural network in each step, by default 32
        preprocessors : list of deoxys.data.Preprocessor, optional
            List of preprocessors to apply on the data, by default None
        x_name : str, optional
            Dataset name to be use as input, by default 'x'
        y_name : str, optional
            Dataset name to be use as target, by default 'y'
        batch_cache : int, optional
            Number of batches to be cached when reading the
            file, by default 10
        train_folds : list of int, or list of str, optional
            List of folds to be use as train data, by default None
        test_folds : list of int, or list of str, optional
            List of folds to be use as test data, by default None
        val_folds : list of int, or list of str, optional
            List of folds to be use as validation data, by default None
        fold_prefix : str, optional
            The prefix of the group name in the HDF5 file, by default 'fold'
    """

    def __init__(self, filename, batch_size=32, preprocessors=None,
                 x_name='x', y_name='y', batch_cache=10,
                 train_folds=None, test_folds=None, val_folds=None,
                 fold_prefix='fold'):
        """
        Initialize a HDF5 Data Reader, which reads data from a HDF5
        file. This file should be split into groups. Each group contain
        datasets, each of which is a column in the data.
        """
        super().__init__()

        h5_filename = file_finder(filename)
        if h5_filename is None:
            # HDF5DataReader is created, but won't be loaded into model
            self.ready = False
            return

        self.hf = h5py.File(h5_filename, 'r')
        self.batch_size = batch_size
        self.batch_cache = batch_cache
        self.preprocessors = preprocessors
        self.x_name = x_name
        self.y_name = y_name
        self.fold_prefix = fold_prefix

        train_folds = list(train_folds) if train_folds else [0]
        test_folds = list(test_folds) if test_folds else [2]
        val_folds = list(val_folds) if val_folds else [1]

        if fold_prefix:
            self.train_folds = ['{}_{}'.format(
                fold_prefix, train_fold) for train_fold in train_folds]
            self.test_folds = ['{}_{}'.format(
                fold_prefix, test_fold) for test_fold in test_folds]
            self.val_folds = ['{}_{}'.format(
                fold_prefix, val_fold) for val_fold in val_folds]
        else:
            self.train_folds = train_folds
            self.test_folds = test_folds
            self.val_folds = val_folds

        self._original_test = None
        self._original_val = None

    @property
    def train_generator(self):
        """

        Returns
        -------
        deoxys.data.DataGenerator
            A DataGenerator for generating batches of data for training
        """
        return HDF5DataGenerator(
            self.hf, batch_size=self.batch_size, batch_cache=self.batch_cache,
            preprocessors=self.preprocessors,
            x_name=self.x_name, y_name=self.y_name,
            folds=self.train_folds)

    @property
    def test_generator(self):
        """

        Returns
        -------
        deoxys.data.DataGenerator
            A DataGenerator for generating batches of data for testing
        """
        return HDF5DataGenerator(
            self.hf, batch_size=self.batch_size, batch_cache=self.batch_cache,
            preprocessors=self.preprocessors,
            x_name=self.x_name, y_name=self.y_name,
            folds=self.test_folds)

    @property
    def val_generator(self):
        """

        Returns
        -------
        deoxys.data.DataGenerator
            A DataGenerator for generating batches of data for validation
        """
        return HDF5DataGenerator(
            self.hf, batch_size=self.batch_size, batch_cache=self.batch_cache,
            preprocessors=self.preprocessors,
            x_name=self.x_name, y_name=self.y_name,
            folds=self.val_folds)

    @property
    def original_test(self):
        """
        Return a dictionary of all data in the test set
        """
        if self._original_test is None:
            self._original_test = {}
            for key in self.hf[self.test_folds[0]].keys():
                data = None
                for fold in self.test_folds:
                    new_data = self.hf[fold][key][:]

                    if data is None:
                        data = new_data
                    else:
                        data = np.concatenate((data, new_data))
                self._original_test[key] = data

        return self._original_test

    @property
    def original_val(self):
        """
        Return a dictionary of all data in the val set
        """
        if self._original_val is None:
            self._original_val = {}
            for key in self.hf[self.val_folds[0]].keys():
                data = None
                for fold in self.val_folds:
                    new_data = self.hf[fold][key][:]

                    if data is None:
                        data = new_data
                    else:
                        data = np.concatenate((data, new_data))
                self._original_val[key] = data

        return self._original_val


class H5Reader(DataReader):
    def __init__(self, filename, batch_size=32, preprocessors=None,
                 x_name='x', y_name='y', batch_cache=10,
                 train_folds=None, test_folds=None, val_folds=None,
                 fold_prefix='fold', shuffle=False, augmentations=None):
        """
        Initialize a HDF5 Data Reader, which reads data from a HDF5
        file. This file should be split into groups. Each group contain
        datasets, each of which is a column in the data.
        """
        super().__init__()

        h5_filename = file_finder(filename)
        if h5_filename is None:
            # HDF5DataReader is created, but won't be loaded into model
            self.ready = False
            return

        self.hf = h5py.File(h5_filename, 'r')
        self.batch_size = batch_size
        self.batch_cache = batch_cache

        self.shuffle = shuffle

        self.preprocessors = preprocessors
        self.augmentations = augmentations

        self.x_name = x_name
        self.y_name = y_name
        self.fold_prefix = fold_prefix

        train_folds = list(train_folds) if train_folds else [0]
        test_folds = list(test_folds) if test_folds else [2]
        val_folds = list(val_folds) if val_folds else [1]

        if fold_prefix:
            self.train_folds = ['{}_{}'.format(
                fold_prefix, train_fold) for train_fold in train_folds]
            self.test_folds = ['{}_{}'.format(
                fold_prefix, test_fold) for test_fold in test_folds]
            self.val_folds = ['{}_{}'.format(
                fold_prefix, val_fold) for val_fold in val_folds]
        else:
            self.train_folds = train_folds
            self.test_folds = test_folds
            self.val_folds = val_folds

        self._original_test = None
        self._original_val = None

    @property
    def train_generator(self):
        """

        Returns
        -------
        deoxys.data.DataGenerator
            A DataGenerator for generating batches of data for training
        """
        return H5DataGenerator(
            self.hf, batch_size=self.batch_size, batch_cache=self.batch_cache,
            preprocessors=self.preprocessors,
            x_name=self.x_name, y_name=self.y_name,
            folds=self.train_folds, shuffle=self.shuffle,
            augmentations=self.augmentations)

    @property
    def test_generator(self):
        """

        Returns
        -------
        deoxys.data.DataGenerator
            A DataGenerator for generating batches of data for testing
        """
        return H5DataGenerator(
            self.hf, batch_size=self.batch_size, batch_cache=self.batch_cache,
            preprocessors=self.preprocessors,
            x_name=self.x_name, y_name=self.y_name,
            folds=self.test_folds, shuffle=False)

    @property
    def val_generator(self):
        """

        Returns
        -------
        deoxys.data.DataGenerator
            A DataGenerator for generating batches of data for validation
        """
        return H5DataGenerator(
            self.hf, batch_size=self.batch_size, batch_cache=self.batch_cache,
            preprocessors=self.preprocessors,
            x_name=self.x_name, y_name=self.y_name,
            folds=self.val_folds, shuffle=False)

    @property
    def original_test(self):
        """
        Return a dictionary of all data in the test set
        """
        if self._original_test is None:
            self._original_test = {}
            for key in self.hf[self.test_folds[0]].keys():
                data = None
                for fold in self.test_folds:
                    new_data = self.hf[fold][key][:]

                    if data is None:
                        data = new_data
                    else:
                        data = np.concatenate((data, new_data))
                self._original_test[key] = data

        return self._original_test

    @property
    def original_val(self):
        """
        Return a dictionary of all data in the val set
        """
        if self._original_val is None:
            self._original_val = {}
            for key in self.hf[self.val_folds[0]].keys():
                data = None
                for fold in self.val_folds:
                    new_data = self.hf[fold][key][:]

                    if data is None:
                        data = new_data
                    else:
                        data = np.concatenate((data, new_data))
                self._original_val[key] = data

        return self._original_val


class DataReaders(metaclass=Singleton):
    """
    A singleton that contains all the registered customized DataReaders
    """

    def __init__(self):
        self._dataReaders = {
            'HDF5Reader': HDF5Reader,
            'H5Reader': H5Reader
        }

    def register(self, key, dr):
        if not issubclass(dr, DataReader):
            raise ValueError(
                "The customized data reader has to be a subclass"
                + " of deoxys.data.DataReader"
            )

        if key in self._dataReaders:
            raise KeyError(
                "Duplicated key, please use another key for this data reader"
            )
        else:
            self._dataReaders[key] = dr

    def unregister(self, key):
        if key in self._dataReaders:
            del self._dataReaders[key]

    @property
    def data_readers(self):
        return self._dataReaders


def register_datareader(key, dr):
    """Register the customized data reader.
    If the key name is already registered, it will raise a KeyError exception.

    Parameters
    ----------
    key : str
        The unique key-name of the data reader
    dr : deoxys.data.DataReader
        The customized data reader class
    """
    DataReaders().register(key, dr)


def unregister_datareader(key):
    """
    Remove the registered data reader with the key-name

    Parameters
    ----------
    key : str
        The key-name of the data reader to be removed
    """
    DataReaders().unregister(key)


def _deserialize(config, custom_objects={}):
    return custom_objects[config['class_name']](**config['config'])


def datareader_from_config(config):
    if 'class_name' not in config:
        raise ValueError('class_name is needed to define data reader')

    if 'config' not in config:
        # auto add empty config for data reader with only class_name
        config['config'] = {}
    return _deserialize(config, custom_objects=DataReaders().data_readers)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import numpy as np


class DataGenerator:
    def __init__(self):
        pass

    @property
    def total_batch(self):
        """
        Total number of batches to iterate all data

        Returns
        -------
        int
            Total number of batches to iterate all data
        """
        return 0

    def generate(self):
        raise NotImplementedError()


class HDF5DataGenerator(DataGenerator):
    def __init__(self, h5file, batch_size=32, batch_cache=10,
                 preprocessors=None,
                 x_name='x', y_name='y', folds=None):
        if not folds or not h5file:
            raise ValueError("h5file or folds is empty")

        # Checking for existence of folds and dataset
        group_names = h5file.keys()
        dataset_names = []
        str_folds = [str(fold) for fold in folds]
        for fold in str_folds:
            if fold not in group_names:
                raise RuntimeError(
                    'HDF5 file: Fold name "{0}" is not in this h5 file'
                    .format(fold))
            if dataset_names:
                if h5file[fold].keys() != dataset_names:
                    raise RuntimeError(
                        'HDF5 file: All folds should have the same structure')
            else:
                dataset_names = h5file[fold].keys()
                if x_name not in dataset_names or y_name not in dataset_names:
                    raise RuntimeError(
                        'HDF5 file: {0} or {1} is not in the file'
                        .format(x_name, y_name))

        # Checking for valida preprocessor
        if preprocessors:
            if type(preprocessors) == list:
                for pp in preprocessors:
                    if not callable(getattr(pp, 'transform', None)):
                        raise ValueError(
                            'Preprocessor should have a "transform" method')
            else:
                if not callable(getattr(preprocessors, 'transform', None)):
                    raise ValueError(
                        'Preprocessor should have a "transform" method')

        self.hf = h5file
        self.batch_size = batch_size
        self.seg_size = batch_size * batch_cache
        self.preprocessors = preprocessors
        self.x_name = x_name
        self.y_name = y_name

        self.folds = str_folds

        # Cache first segment for first fold
        self.index = 0
        self.seg_index = 0

        first_fold_name = self.folds[0]

        self.x_cur = self.hf[first_fold_name][self.x_name][:self.seg_size]
        self.y_cur = self.hf[first_fold_name][self.y_name][:self.seg_size]

        if self.preprocessors:
            if type(self.preprocessors) == list:
                for preprocessor in self.preprocessors:
                    self.x_cur, self.y_cur = preprocessor.transform(
                        self.x_cur, self.y_cur)
            else:
                self.x_cur, self.y_cur = self.preprocessors.transform(
                    self.x_cur, self.y_cur)

        # Get the total length of the first fold
        self.fold_len = len(self.hf[first_fold_name][self.y_name])

        self._total_batch = None

    @property
    def total_batch(self):
        """Total number of batches to iterate all data.
        It will be used as the number of steps per epochs when training or
        validating data in a model.

        Returns
        -------
        int
            Total number of batches to iterate all data
        """
        if self._total_batch is None:
            total_batch = 0
            fold_names = self.folds

            for fold_name in fold_names:
                total_batch += np.ceil(
                    len(self.hf[fold_name][self.y_name]) / self.batch_size)
            self._total_batch = int(total_batch)
        return self._total_batch

    def next_fold(self):
        # Reset segment index
        self.seg_index = 0

        # Remove previous fold index and move to next one
        self.folds.append(self.folds.pop(0))

        fold_name = self.folds[0]
        y = self.hf[fold_name][self.y_name]

        # Recalculate the total length
        self.fold_len = len(y)

    def next_seg(self):
        # Reset index
        self.index = 0

        # Move segment index
        self.seg_index += self.seg_size

        # When all segments fold has been yielded, move to next fold
        if self.seg_index >= self.fold_len:
            self.next_fold()

        # store local variable after seg_index changed
        seg_index = self.seg_index

        fold_name = self.folds[0]
        # The last segment may has less items than seg_size
        if seg_index + self.seg_size >= self.fold_len:
            self.x_cur = self.hf[fold_name][self.x_name][seg_index:]
            self.y_cur = self.hf[fold_name][self.y_name][seg_index:]
        else:
            next_seg_index = seg_index + self.seg_size

            self.x_cur = self.hf[fold_name][
                self.x_name][seg_index:next_seg_index]
            self.y_cur = self.hf[fold_name][
                self.y_name][seg_index:next_seg_index]

        # Apply preprocessor
        if self.preprocessors:
            if type(self.preprocessors) == list:
                for preprocessor in self.preprocessors:
                    self.x_cur, self.y_cur = preprocessor.transform(
                        self.x_cur, self.y_cur)
            else:
                self.x_cur, self.y_cur = self.preprocessors.transform(
                    self.x_cur, self.y_cur)

    def generate(self):
        """Create a generator that generate a batch of data

        Yields
        -------
        tuple of 2 arrays
            batch of (input, target)
        """
        while True:
            # When all batches of data are yielded, move to next seg
            if self.index >= self.seg_size or \
                    self.seg_index + self.index >= self.fold_len:
                self.next_seg()

            # Index may has been reset. Thus, call after next_seg
            index = self.index

            # The last batch of data may not have less than batch_size items
            if index + self.batch_size >= self.seg_size or \
                    self.seg_index + index + self.batch_size >= self.fold_len:
                batch_x = self.x_cur[index:]
                batch_y = self.y_cur[index:]
            else:
                # Take the next batch
                batch_x = self.x_cur[index:(index + self.batch_size)]
                batch_y = self.y_cur[index:(index + self.batch_size)]

            self.index += self.batch_size
            yield batch_x, batch_y


class H5DataGenerator(DataGenerator):
    def __init__(self, h5file, batch_size=32, batch_cache=10,
                 preprocessors=None,
                 x_name='x', y_name='y', folds=None,
                 shuffle=False, augmentations=None):
        if not folds or not h5file:
            raise ValueError("h5file or folds is empty")

        # Checking for existence of folds and dataset
        group_names = h5file.keys()
        dataset_names = []
        str_folds = [str(fold) for fold in folds]
        for fold in str_folds:
            if fold not in group_names:
                raise RuntimeError(
                    'HDF5 file: Fold name "{0}" is not in this h5 file'
                    .format(fold))
            if dataset_names:
                if h5file[fold].keys() != dataset_names:
                    raise RuntimeError(
                        'HDF5 file: All folds should have the same structure')
            else:
                dataset_names = h5file[fold].keys()
                if x_name not in dataset_names or y_name not in dataset_names:
                    raise RuntimeError(
                        'HDF5 file: {0} or {1} is not in the file'
                        .format(x_name, y_name))

        # Checking for valid preprocessor
        if preprocessors:
            if type(preprocessors) == list:
                for pp in preprocessors:
                    if not callable(getattr(pp, 'transform', None)):
                        raise ValueError(
                            'Preprocessor should have a "transform" method')
            else:
                if not callable(getattr(preprocessors, 'transform', None)):
                    raise ValueError(
                        'Preprocessor should have a "transform" method')

        if augmentations:
            if type(augmentations) == list:
                for pp in augmentations:
                    if not callable(getattr(pp, 'transform', None)):
                        raise ValueError(
                            'Augmentation must be a preprocessor with'
                            ' a "transform" method')
            else:
                if not callable(getattr(augmentations, 'transform', None)):
                    raise ValueError(
                        'Augmentation must be a preprocessor with'
                        ' a "transform" method')

        self.hf = h5file
        self.batch_size = batch_size
        self.seg_size = batch_size * batch_cache
        self.preprocessors = preprocessors
        self.augmentations = augmentations

        self.x_name = x_name
        self.y_name = y_name

        self.shuffle = shuffle

        self.folds = str_folds

        self._total_batch = None

        # initialize "index" of current seg and fold
        self.seg_idx = 0
        self.fold_idx = 0

        # shuffle the folds
        if self.shuffle:
            np.random.shuffle(self.folds)

        # calculate number of segs in this fold
        seg_num = np.ceil(
            h5file[self.folds[0]][y_name].shape[0] / self.seg_size)

        self.seg_list = np.arange(seg_num).astype(int)
        if self.shuffle:
            np.random.shuffle(self.seg_list)

    @property
    def total_batch(self):
        """Total number of batches to iterate all data.
        It will be used as the number of steps per epochs when training or
        validating data in a model.

        Returns
        -------
        int
            Total number of batches to iterate all data
        """
        if self._total_batch is None:
            total_batch = 0
            fold_names = self.folds

            for fold_name in fold_names:
                total_batch += np.ceil(
                    len(self.hf[fold_name][self.y_name]) / self.batch_size)
            self._total_batch = int(total_batch)
        return self._total_batch

    def next_fold(self):
        self.fold_idx += 1

        if self.fold_idx == len(self.folds):
            self.fold_idx = 0

            if self.shuffle:
                np.random.shuffle(self.folds)

    def next_seg(self):
        if self.seg_idx == len(self.seg_list):
            # move to next fold
            self.next_fold()

            # reset seg index
            self.seg_idx = 0
            # recalculate seg_num
            cur_fold = self.folds[self.fold_idx]
            seg_num = np.ceil(
                self.hf[cur_fold][self.y_name].shape[0] / self.seg_size)

            self.seg_list = np.arange(seg_num).astype(int)

            if self.shuffle:
                np.random.shuffle(self.seg_list)

        cur_fold = self.folds[self.fold_idx]
        cur_seg_idx = self.seg_list[self.seg_idx]

        start, end = cur_seg_idx * \
            self.seg_size, (cur_seg_idx + 1) * self.seg_size

        # print(cur_fold, cur_seg_idx, start, end)

        seg_x = self.hf[cur_fold][self.x_name][start: end]
        seg_y = self.hf[cur_fold][self.y_name][start: end]

        return_indice = np.arange(len(seg_y))

        if self.shuffle:
            np.random.shuffle(return_indice)

        # Apply preprocessor
        if self.preprocessors:
            if type(self.preprocessors) == list:
                for preprocessor in self.preprocessors:
                    seg_x, seg_y = preprocessor.transform(
                        seg_x, seg_y)
            else:
                seg_x, seg_y = self.preprocessors.transform(
                    seg_x, seg_y)
        # Apply augmentation:
        if self.augmentations:
            if type(self.augmentations) == list:
                for preprocessor in self.augmentations:
                    seg_x, seg_y = preprocessor.transform(
                        seg_x, seg_y)
            else:
                seg_x, seg_y = self.augmentations.transform(
                    seg_x, seg_y)

        # increase seg index
        self.seg_idx += 1

        return seg_x[return_indice], seg_y[return_indice]

    def generate(self):
        """Create a generator that generate a batch of data

        Yields
        -------
        tuple of 2 arrays
            batch of (input, target)
        """
        while True:
            seg_x, seg_y = self.next_seg()

            seg_len = len(seg_y)

            for i in range(0, seg_len, self.batch_size):
                batch_x = seg_x[i:(i + self.batch_size)]
                batch_y = seg_y[i:(i + self.batch_size)]

                # print(batch_x.shape)

                yield batch_x, batch_y



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import numpy as np
from deoxys_image import normalize, apply_affine_transform, apply_flip
from deoxys_image import ImageAugmentation
from deoxys.keras.preprocessing import ImageDataGenerator
from ..utils import Singleton


class BasePreprocessor:
    def __init__(self, **kwargs):
        """
        Wrapper for creating a trained/fitted Preprocessor instance
        """
        pass

    def transform(self, x, y):
        return x, y


class DummyPreprocessor:
    def __init__(self):
        pass

    def transform(self, x, y):
        return x, y


class SingleChannelPreprocessor(BasePreprocessor):
    """Make single channel images have one axis for the channel
    """

    def transform(self, x, y):
        x_shape = tuple(list(x.shape) + [1])
        y_shape = tuple(list(y.shape) + [1])

        return x.reshape(x_shape), y.reshape(y_shape)


class WindowingPreprocessor(BasePreprocessor):
    """Set the range of the images

        Parameters
        ----------
        window_center : int, float
            the center of the range
        window_width : int, float
            the range
        channel : int
            the index of the channel to apply windowing
    """

    def __init__(self, window_center, window_width, channel):
        self.window_center, self.window_width = window_center, window_width
        self.channel = channel

    def perform_windowing(self, image):
        image = image - self.window_center
        image[image < -self.window_width / 2] = -self.window_width / 2
        image[image > self.window_width / 2] = self.window_width / 2
        return image

    def transform(self, images, targets):
        images = images.copy()
        images[..., self.channel] = self.perform_windowing(
            images[..., self.channel])
        return images, targets


class HounsfieldWindowingPreprocessor(WindowingPreprocessor):
    """Set the range of the images

        Parameters
        ----------
        window_center : int, float
            the center of the range
        window_width : int, float
            the range
        channel : int
            the index of the channel to apply windowing
        hounsfield_offset: int, float
            the Hounsfield offset
    """

    def __init__(self, window_center, window_width, channel,
                 hounsfield_offset=1024):
        super().__init__(
            window_center+hounsfield_offset, window_width, channel)


class ImageNormalizerPreprocessor(BasePreprocessor):
    """
        Normalize all channels to the range of the close interval [0, 1]

        Parameters
        ----------
        vmin : int, float, list, tuple, optional.
            If an int or a float, it will be the lower limits in all channels,
            else it should be a list of lower values associated with all axes.
            By default None (choosing the minimum value of
            each channel in the image batch)
        vmax : int, float, list, tuple, optional
            If an int or a float, it will be the upper limits in all channels,
            else it should be a list of upper values associated with all axes.
            By default None (choosing the maximum value of
            each channel in the image batch)
    """

    def __init__(self, vmin=None, vmax=None):
        self.vmin = vmin
        self.vmax = vmax

    def transform(self, images, targets):
        transformed_images = normalize(images, self.vmin, self.vmax)

        return transformed_images, targets


class ChannelRemoval(BasePreprocessor):
    """Remove one or more channels from the images

        Parameters
        ----------
        channel : int, list, tuple, optional
            the index of the channel to be removed, by default 1
    """

    def __init__(self, channel=1):
        self.channel = channel

    def transform(self, images, targets):
        return np.delete(images, self.channel, axis=-1), targets


class ChannelSelector(BasePreprocessor):
    """Select / filter one or more channels from the images

        Parameters
        ----------
        channel : int, list, tuple, optional
            the index of the channel to be selected, by default 0
    """

    def __init__(self, channel=0):
        if '__iter__' not in dir(channel):
            self.channel = [channel]
        else:
            self.channel = channel

    def transform(self, images, targets):
        remove_channel = [c for c in np.arange(
            images.shape[-1]) if c not in self.channel]
        return np.delete(images, remove_channel, axis=-1), targets


class UnetPaddingPreprocessor(BasePreprocessor):
    """
        Pad the images so that their sizes would not change after going
        through a standard Unet model.

        Parameters
        ----------
        depth : int, optional
            number of maxpooling layer in the Unet, by default 4
        mode : str, optional
            way to fill the values in the paddings,
            currently only support 'auto' (pad with 0)
    """

    def __init__(self, depth=4, mode='auto'):
        self.depth = depth
        self.mode = mode

    def transform(self, images, targets):
        image_shape = images.shape
        target_shape = targets.shape

        shape = image_shape[1:-1]

        divide_factor = 2 ** self.depth

        if len(shape) == 2:
            height, width = shape
            if height % divide_factor != 0:
                new_height = (height // divide_factor + 1) * divide_factor
            else:
                new_height = height

            if width % divide_factor != 0:
                new_width = (width // divide_factor + 1) * divide_factor
            else:
                new_width = width

            new_images = np.zeros(
                (image_shape[0],
                 new_height, new_width,
                 image_shape[-1]))
            new_targets = np.zeros(
                (target_shape[0],
                    new_height, new_width,
                    target_shape[-1]))

            min_h = (new_height - height) // 2
            min_w = (new_width - width) // 2

            new_images[:, min_h: min_h+height,
                       min_w: min_w+width, :] = images

            new_targets[:, min_h: min_h+height,
                        min_w: min_w+width, :] = targets

            return new_images, new_targets

        if len(shape) == 3:
            height, width, z = shape

            if height % divide_factor != 0:
                new_height = (height // divide_factor + 1) * divide_factor
            else:
                new_height = height

            if width % divide_factor != 0:
                new_width = (width // divide_factor + 1) * divide_factor
            else:
                new_width = width

            if z % divide_factor != 0:
                new_z = (z // divide_factor + 1) * divide_factor
            else:
                new_z = z

            if self.mode == 'edge':
                pass
            else:  # default - pad with zeros
                new_images = np.zeros(
                    (image_shape[0],
                     new_height, new_width, new_z,
                     image_shape[-1]))
                new_targets = np.zeros(
                    (target_shape[0],
                     new_height, new_width, new_z,
                     target_shape[-1]))

            min_h = (new_height - height) // 2
            min_w = (new_width - width) // 2
            min_z = (new_z - z) // 2

            new_images[:, min_h: min_h+height,
                       min_w: min_w+width, min_z:min_z+z, :] = images

            new_targets[:, min_h: min_h+height,
                        min_w: min_w+width, min_z:min_z+z, :] = targets

            return new_images, new_targets

        raise RuntimeError('Does not support 4D tensors')


class ImageAffineTransformPreprocessor(BasePreprocessor):
    def __init__(self, rotation_degree=0, rotation_axis=2, zoom_factor=1,
                 shift=None, flip_axis=None,
                 fill_mode='constant', cval=0):
        """Apply affine transformation to all images

        Parameters
        ----------
        rotation_degree : int, optional
            the degree to rotate, by default 0
        rotation_axis : int, optional
            the rotation axis, by default 2
        zoom_factor : int, optional
            , by default 1
        shift : list, tuple, optional
            the pixels of shifting in each axis, by default None
        flip_axis : int, optional
            the axis to be flipped, by default None
        fill_mode : str, optional
            the mode to fill when applying affine transformation,
            by default 'constant'
        cval : int, optional
            by default 0
        """

        self.affine_transform = rotation_degree > 0 or \
            zoom_factor != 1 or shift is not None

        self.rotation_degree = rotation_degree
        self.rotation_axis = rotation_axis
        self.zoom_factor = zoom_factor
        self.shift = shift
        self.flip_axis = flip_axis
        self.fill_mode = fill_mode
        self.cval = cval

    def transform(self, images, targets):
        transformed_images = images.copy()
        transformed_targets = targets.copy()

        # loop through
        for i in range(len(images)):
            # apply affine transform if possible
            if self.affine_transform:
                # After affine transform, the pixel intensity may change
                # the image should clip back to original range
                reduced_ax = tuple(
                    range(len(transformed_images[i].shape) - 1))
                vmin = transformed_images[i].min(axis=reduced_ax)
                vmax = transformed_images[i].max(axis=reduced_ax)

                transformed_images[i] = apply_affine_transform(
                    transformed_images[i],
                    mode=self.fill_mode, cval=self.cval,
                    theta=self.rotation_degree,
                    rotation_axis=self.rotation_axis,
                    zoom_factor=self.zoom_factor,
                    shift=self.shift).clip(vmin, vmax)

                transformed_targets[i] = apply_affine_transform(
                    transformed_targets[i],
                    mode=self.fill_mode, cval=self.cval,
                    theta=self.rotation_degree,
                    rotation_axis=self.rotation_axis,
                    zoom_factor=self.zoom_factor,
                    shift=self.shift)

                # round the target label back to integer
                transformed_targets[i] = np.rint(
                    transformed_targets[i])

            # flip image
            if self.flip_axis is not None:
                transformed_images[i] = apply_flip(
                    transformed_images[i], self.flip_axis)

                transformed_targets[i] = apply_flip(
                    transformed_targets[i], self.flip_axis)

        return transformed_images, transformed_targets


class ImageAugmentation2D(BasePreprocessor):
    r"""
        Apply transformation in 2d image (and mask label) for augmentation.
        Check `ImageAugmentation3D` for augmentation on 3d images

        Parameters
        ----------
        rotation_range : int, optional
            range of the angle rotation, in degree, by default 0 (no rotation)
        rotation_chance : float, optional
            probability to apply rotation transformation to an image,
            by default 0.2
        zoom_range : float, list, tuple optional
            the range of zooming, zooming in when the number is less than 1,
            and zoom out when the number if larger than 1.
            If a `float`, then it is the range between that number and 1,
            by default 1 (no zooming)
        zoom_chance : float, optional
            probability to apply zoom transformation to an image,
            by default 0.2
        shift_range : tuple or list, optional
            the range of translation in each axis, by default None (no shifts)
        shift_chance : float, optional
            probability to apply translation transformation to an image,
            by default 0.1
        flip_axis : int, tuple, list, optional
            flip by one or more axis (in the single image) with a probability
            of 0.5, by default None (no flipping).
            `flip_axis=0` means the image will be flipped vertically, while
            `flip_axis=1` means the image will be flipped horizontally.
        brightness_range : int, tuple, list, optional
            range of the brightness portion,
            based on the max intensity value of each channel.
            For example, when the max intensity value of one channel is 1.0,
            and the brightness is chaned by 1.2, then every pixel in that
            channel will increase the intensity value by 0.2.

            .. math:: 0.2 = 1.0 \cdot (1.2 - 1)

            By default 1 (no changes in brightness)
        brightness_channel : int, tuple, list, optional
            the channel(s) to apply changes in brightness,
            by default None (apply to all channels)
        brightness_chance : float, optional
            probability to apply brightness change transform to an image,
            by default 0.1
        contrast_range : int, tuple, list, optional
            range of the contrast portion,
            (the history range is scaled up or down).
            By default 1 (no changes in contrast)
        contrast_channel : int, tuple, list, optional
            the channel(s) to apply changes in contrast,
            by default None (apply to all channels)
        contrast_chance : float, optional
            probability to apply contrast change transform to an image,
            by default 0.1
        noise_variance : float, tuple, list, optional
            range of the noise variance
            when adding Gaussian noise to the image,
            by default 0 (no adding noise)
        noise_channel : int, tuple, list, optional
            the channel(s) to apply Gaussian noise,
            by default None (apply to all channels)
        noise_chance : float, optional
            probability to apply gaussian noise to an image,
            by default 0.1
        blur_range : int, tuple, list, optional
            range of the blur sigma
            when applying the Gaussian filter to the image,
            by default 0 (no blur)
        blur_channel :int, tuple, list, optional
            the channel(s) to apply Gaussian blur,
            by default None (apply to all channels)
        blur_chance : float, optional
            probability to apply gaussian blur to an image,
            by default 0.1
        fill_mode : str, optional
            the fill mode in affine transformation
            (rotation, zooming, shifting / translation),
            one of {'reflect', 'constant', 'nearest', 'mirror', 'wrap'},
            by default 'constant'
        cval : int, optional
            When rotation, or zooming, or shifting is applied to the image,
            `cval` is the value to fill past edges of input
            if `fill_mode` is 'constant'.
            By default 0
        """

    _RANK = 3
    _ROTATION_AXIS = 2

    def __init__(self, rotation_range=0, rotation_chance=0.2,
                 zoom_range=1, zoom_chance=0.2,
                 shift_range=None, shift_chance=0.1,
                 flip_axis=None,
                 brightness_range=1, brightness_channel=None,
                 brightness_chance=0.1,
                 contrast_range=1, contrast_channel=None,
                 contrast_chance=0.1,
                 noise_variance=0, noise_channel=None,
                 noise_chance=0.1,
                 blur_range=0, blur_channel=None, blur_chance=0.1,
                 fill_mode='constant', cval=0):
        self.augmentation_obj = ImageAugmentation(
            self._RANK,
            rotation_range, self._ROTATION_AXIS, rotation_chance,
            zoom_range, zoom_chance,
            shift_range, shift_chance,
            flip_axis,
            brightness_range, brightness_channel,
            brightness_chance,
            contrast_range, contrast_channel,
            contrast_chance,
            noise_variance, noise_channel,
            noise_chance,
            blur_range, blur_channel, blur_chance,
            fill_mode, cval
        )

    def transform(self, images, targets):
        """
        Apply augmentation to a batch of images

        Parameters
        ----------
        images : np.array
            the image batch
        targets : np.array, optional
            the target batch, by default None

        Returns
        -------
        np.array
            the transformed images batch (and target)
        """
        return self.augmentation_obj.transform(images, targets)


class ImageAugmentation3D(ImageAugmentation2D):
    r"""
        Apply transformation in 3d image (and mask label) for augmentation

        Parameters
        ----------
        rotation_range : int, optional
            range of the angle rotation, in degree, by default 0 (no rotation)
        rotation_axis : int, optional
            the axis of one image to apply rotation,
            by default 0
        rotation_chance : float, optional
            probability to apply rotation transformation to an image,
            by default 0.2
        zoom_range : float, list, tuple optional
            the range of zooming, zooming in when the number is less than 1,
            and zoom out when the number if larger than 1.
            If a `float`, then it is the range between that number and 1,
            by default 1 (no zooming)
        zoom_chance : float, optional
            probability to apply zoom transformation to an image,
            by default 0.2
        shift_range : tuple or list, optional
            the range of translation in each axis, by default None (no shifts)
        shift_chance : float, optional
            probability to apply translation transformation to an image,
            by default 0.1
        flip_axis : int, tuple, list, optional
            flip by one or more axis (in the single image),
            by default None (no flipping)
        brightness_range : int, tuple, list, optional
            range of the brightness portion,
            based on the max intensity value of each channel.
            For example, when the max intensity value of one channel is 1.0,
            and the brightness is chaned by 1.2, then every pixel in that
            channel will increase the intensity value by 0.2.

            .. math:: 0.2 = 1.0 \cdot (1.2 - 1)

            By default 1 (no changes in brightness)
        brightness_channel : int, tuple, list, optional
            the channel(s) to apply changes in brightness,
            by default None (apply to all channels)
        brightness_chance : float, optional
            probability to apply brightness change transform to an image,
            by default 0.1
        contrast_range : int, tuple, list, optional
            range of the contrast portion,
            (the history range is scaled up or down).
            By default 1 (no changes in contrast)
        contrast_channel : int, tuple, list, optional
            the channel(s) to apply changes in contrast,
            by default None (apply to all channels)
        contrast_chance : float, optional
            probability to apply contrast change transform to an image,
            by default 0.1
        noise_variance : float, tuple, list, optional
            range of the noise variance
            when adding Gaussian noise to the image,
            by default 0 (no adding noise)
        noise_channel : int, tuple, list, optional
            the channel(s) to apply Gaussian noise,
            by default None (apply to all channels)
        noise_chance : float, optional
            probability to apply gaussian noise to an image,
            by default 0.1
        blur_range : int, tuple, list, optional
            range of the blur sigma
            when applying the Gaussian filter to the image,
            by default 0 (no blur)
        blur_channel :int, tuple, list, optional
            the channel(s) to apply Gaussian blur,
            by default None (apply to all channels)
        blur_chance : float, optional
            probability to apply gaussian blur to an image,
            by default 0.1
        fill_mode : str, optional
            the fill mode in affine transformation
            (rotation, zooming, shifting / translation),
            one of {'reflect', 'constant', 'nearest', 'mirror', 'wrap'},
            by default 'constant'
        cval : int, optional
            When rotation, or zooming, or shifting is applied to the image,
            `cval` is the value to fill past edges of input
            if `fill_mode` is 'constant'.
            By default 0
        """
    _RANK = 4

    def __init__(self, rotation_range=0, rotation_axis=0, rotation_chance=0.2,
                 zoom_range=1, zoom_chance=0.2,
                 shift_range=None, shift_chance=0.1,
                 flip_axis=None,
                 brightness_range=1, brightness_channel=None,
                 brightness_chance=0.1,
                 contrast_range=1, contrast_channel=None,
                 contrast_chance=0.1,
                 noise_variance=0, noise_channel=None,
                 noise_chance=0.1,
                 blur_range=0, blur_channel=None, blur_chance=0.1,
                 fill_mode='constant', cval=0):
        self.augmentation_obj = ImageAugmentation(
            self._RANK,
            rotation_range, rotation_axis, rotation_chance,
            zoom_range, zoom_chance,
            shift_range, shift_chance,
            flip_axis,
            brightness_range, brightness_channel,
            brightness_chance,
            contrast_range, contrast_channel,
            contrast_chance,
            noise_variance, noise_channel,
            noise_chance,
            blur_range, blur_channel, blur_chance,
            fill_mode, cval
        )


class KerasImagePreprocessorX(BasePreprocessor):
    """Apply keras image augmentation to the input images
    """

    def __init__(self,
                 shuffle=False,
                 featurewise_center=False,
                 samplewise_center=False,
                 featurewise_std_normalization=False,
                 samplewise_std_normalization=False,
                 zca_whitening=False, zca_epsilon=1e-06,
                 rotation_range=0, width_shift_range=0.0,
                 height_shift_range=0.0,
                 brightness_range=None, shear_range=0.0, zoom_range=0.0,
                 channel_shift_range=0.0, fill_mode='nearest', cval=0.0,
                 horizontal_flip=False, vertical_flip=False, rescale=None,
                 scale_down=None,
                 preprocessing_function=None, data_format='channels_last',
                 interpolation_order=1, dtype='float32'):

        self.shuffle = shuffle

        if scale_down and (rescale is None):
            rescale = 1 / scale_down

        self.preprocessor = ImageDataGenerator(
            featurewise_center=featurewise_center,
            samplewise_center=samplewise_center,
            featurewise_std_normalization=featurewise_std_normalization,
            samplewise_std_normalization=samplewise_std_normalization,
            zca_whitening=zca_whitening, zca_epsilon=zca_epsilon,
            rotation_range=rotation_range, width_shift_range=width_shift_range,
            height_shift_range=height_shift_range,
            brightness_range=brightness_range, shear_range=shear_range,
            zoom_range=zoom_range,
            channel_shift_range=channel_shift_range, fill_mode=fill_mode,
            cval=cval,
            horizontal_flip=horizontal_flip, vertical_flip=vertical_flip,
            rescale=rescale,
            preprocessing_function=preprocessing_function,
            data_format=data_format, dtype=dtype)

    def transform(self, x, y):
        return next(self.preprocessor.flow(x, batch_size=x.shape[0],
                                           shuffle=self.shuffle)), y


class KerasImagePreprocessorY(BasePreprocessor):
    def __init__(self,
                 shuffle=True,
                 featurewise_center=False,
                 samplewise_center=False,
                 featurewise_std_normalization=False,
                 samplewise_std_normalization=False,
                 zca_whitening=False, zca_epsilon=1e-06,
                 rotation_range=0, width_shift_range=0.0,
                 height_shift_range=0.0,
                 brightness_range=None, shear_range=0.0, zoom_range=0.0,
                 channel_shift_range=0.0, fill_mode='nearest', cval=0.0,
                 horizontal_flip=False, vertical_flip=False, rescale=None,
                 scale_down=None,
                 preprocessing_function=None, data_format='channels_last',
                 dtype='float32'):
        self.shuffle = shuffle

        if scale_down and (rescale is None):
            rescale = 1 / scale_down

        self.preprocessor = ImageDataGenerator(
            featurewise_center=featurewise_center,
            samplewise_center=samplewise_center,
            featurewise_std_normalization=featurewise_std_normalization,
            samplewise_std_normalization=samplewise_std_normalization,
            zca_whitening=zca_whitening, zca_epsilon=zca_epsilon,
            rotation_range=rotation_range, width_shift_range=width_shift_range,
            height_shift_range=height_shift_range,
            brightness_range=brightness_range, shear_range=shear_range,
            zoom_range=zoom_range,
            channel_shift_range=channel_shift_range, fill_mode=fill_mode,
            cval=cval,
            horizontal_flip=horizontal_flip, vertical_flip=vertical_flip,
            rescale=rescale,
            preprocessing_function=preprocessing_function,
            data_format=data_format,
            dtype=dtype)

    def transform(self, x, y):
        return x, next(self.preprocessor.flow(y, batch_size=x.shape[0],
                                              shuffle=self.shuffle))


class Preprocessors(metaclass=Singleton):
    """
    A singleton that contains all the registered customized preprocessors
    """

    def __init__(self):
        self._preprocessors = {
            'WindowingPreprocessor': WindowingPreprocessor,
            'HounsfieldWindowingPreprocessor': HounsfieldWindowingPreprocessor,
            'ImageNormalizerPreprocessor': ImageNormalizerPreprocessor,
            'UnetPaddingPreprocessor': UnetPaddingPreprocessor,
            'ChannelSelector': ChannelSelector,
            'ChannelRemoval': ChannelRemoval,
            'ImageAffineTransformPreprocessor':
                ImageAffineTransformPreprocessor,
            'ImageAugmentation2D': ImageAugmentation2D,
            'ImageAugmentation3D': ImageAugmentation3D,
            'SingleChannelPreprocessor': SingleChannelPreprocessor,
            'KerasImagePreprocessorX': KerasImagePreprocessorX,
            'KerasImagePreprocessorY': KerasImagePreprocessorY
        }

    def register(self, key, preprocessor):
        if not issubclass(preprocessor, BasePreprocessor):
            raise ValueError(
                "The customized preprocessor has to be a subclass"
                + " of deoxys.data.BasePreprocessor"
            )

        if key in self._preprocessors:
            raise KeyError(
                "Duplicated key, please use another key for this preprocessor"
            )
        else:
            self._preprocessors[key] = preprocessor

    def unregister(self, key):
        if key in self._preprocessors:
            del self._preprocessors[key]

    @property
    def preprocessors(self):
        return self._preprocessors


def register_preprocessor(key, preprocessor):
    """
    Register the customized preprocessor.
    If the key name is already registered, it will raise a KeyError exception

    Parameters
    ----------
    key : str
        The unique key-name of the preprocessor
    preprocessor : deoxys.data.BasePreprocessor
        The customized preprocessor class
    """
    Preprocessors().register(key, preprocessor)


def unregister_preprocessor(key):
    """
    Remove the registered preprocessor with the key-name

    Parameters
    ----------
    key : str
        The key-name of the preprocessor to be removed
    """
    Preprocessors().unregister(key)


def _deserialize(config, custom_objects={}):
    predefined_obj = {
        'DummyPreprocessor': DummyPreprocessor
    }

    predefined_obj.update(custom_objects)

    return predefined_obj[config['class_name']](**config['config'])


def preprocessor_from_config(config):
    if 'class_name' not in config:
        raise ValueError('class_name is needed to define preprocessor')

    if 'config' not in config:
        # auto add empty config for preprocessor with only class_name
        config['config'] = {}
    return _deserialize(config, custom_objects=Preprocessors().preprocessors)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest
import numpy as np
from deoxys.keras import backend as K
from deoxys.keras.losses import Loss
from deoxys.model.losses import Losses, BinaryFbetaLoss, ModifiedDiceLoss
from deoxys.customize import register_loss, \
    unregister_loss, custom_loss
from deoxys.utils import Singleton


@pytest.fixture(autouse=True)
def clear_singleton():
    Singleton._instances = {}  # clear singleton


@pytest.fixture
def loss_class():
    class TestLoss(Loss):
        pass

    yield TestLoss


def test_is_singleton():
    losses_instance = Losses()
    another_instance = Losses()

    assert losses_instance is another_instance


def test_register_random_obj():
    with pytest.raises(ValueError):
        register_loss("TestLoss", object)


def test_register_loss_success(loss_class):
    register_loss("TestLoss", loss_class)

    assert Losses()._losses["TestLoss"] is loss_class


def test_register_duplicate_loss(loss_class):
    register_loss("TestLoss", loss_class)

    with pytest.raises(KeyError):
        register_loss("TestLoss", loss_class)


def test_unregister_loss(loss_class):
    register_loss("TestLoss", loss_class)

    assert Losses()._losses["TestLoss"] is loss_class
    unregister_loss("TestLoss")

    assert "TestLoss" not in Losses()._losses


def test_decorator():
    @custom_loss
    class TestLoss2(Loss):
        pass

    assert Losses()._losses["TestLoss2"] is TestLoss2


def test_binary_fbeta_loss():
    true = [
        [
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0]
        ],
        [
            [0, 1, 0],
            [1, 1, 1],
            [0, 1, 0]
        ],
        [
            [1, 1, 1],
            [0, 1, 0],
            [1, 1, 1]
        ],
        [
            [1, 0, 1],
            [0, 1, 0],
            [1, 0, 1]
        ]
    ]

    pred = [
        [
            [0, 0, 0],
            [0, 1, 0],
            [0, 1, 0]
        ],
        [
            [0, 1, 0],
            [1, 1, 1],
            [0, 0, 0]
        ],
        [
            [1, 1, 1],
            [0, 0, 1],
            [1, 1, 1]
        ],
        [
            [1, 1, 1],
            [0, 1, 0],
            [1, 0, 1]
        ]
    ]

    true = K.constant(true)
    pred = K.constant(pred)
    tp = K.constant([2, 4, 6, 5])
    fp = K.constant([0, 0, 1, 1])
    fn = K.constant([1, 1, 1, 0])

    eps = 1e-8

    def fscore(beta, tp, fp, fn):
        numerator = (1 + beta**2) * tp + eps
        denominator = (1 + beta**2) * tp + beta**2 * fn + fp + eps

        return numerator / denominator

    beta = 1
    loss = BinaryFbetaLoss(beta=beta)
    assert np.allclose(loss.call(true, pred), (1 - fscore(beta, tp, fp, fn)))

    beta = 2
    loss = BinaryFbetaLoss(beta=beta)
    assert np.allclose(loss.call(true, pred), (1 - fscore(beta, tp, fp, fn)))


def test_dice_loss():
    true = [
        [
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0]
        ],
        [
            [0, 1, 0],
            [1, 1, 1],
            [0, 1, 0]
        ],
        [
            [1, 1, 1],
            [0, 1, 0],
            [1, 1, 1]
        ],
        [
            [1, 0, 1],
            [0, 1, 0],
            [1, 0, 1]
        ]
    ]

    pred = [
        [
            [0, 0, 0],
            [0, 1, 0],
            [0, 1, 0]
        ],
        [
            [0, 1, 0],
            [1, 1, 1],
            [0, 0, 0]
        ],
        [
            [1, 1, 1],
            [0, 0, 1],
            [1, 1, 1]
        ],
        [
            [1, 1, 1],
            [0, 1, 0],
            [1, 0, 1]
        ]
    ]

    true = K.constant(true)
    pred = K.constant(pred)
    tp = K.constant([2, 4, 6, 5])
    fp = K.constant([0, 0, 1, 1])
    fn = K.constant([1, 1, 1, 0])

    eps = 1e-8

    def fscore(beta, tp, fp, fn):
        numerator = (1 + beta**2) * tp + eps
        denominator = (1 + beta**2) * tp + beta**2 * fn + fp + eps

        return numerator / denominator

    beta = 1
    loss = ModifiedDiceLoss(beta=beta)
    assert np.allclose(loss.call(true, pred), (1 - fscore(beta, tp, fp, fn)))

    beta = 2
    loss = ModifiedDiceLoss(beta=beta)
    assert np.allclose(loss.call(true, pred), (1 - fscore(beta, tp, fp, fn)))



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest
import numpy as np
from deoxys.keras import backend as K
from deoxys.keras.metrics import Metric
from deoxys.model.metrics import Metrics, BinaryFbeta, Fbeta, Dice
from deoxys.customize import register_metric, \
    unregister_metric, custom_metric
from deoxys.utils import Singleton


@pytest.fixture(autouse=True)
def clear_singleton():
    Singleton._instances = {}  # clear singleton


@pytest.fixture
def metric_class():
    class TestMetric(Metric):
        pass

    yield TestMetric


def test_is_singleton():
    metrics_instance = Metrics()
    another_instance = Metrics()

    assert metrics_instance is another_instance


def test_register_random_obj():
    with pytest.raises(ValueError):
        register_metric("TestMetric", object)


def test_register_metric_success(metric_class):
    register_metric("TestMetric", metric_class)

    assert Metrics()._metrics["TestMetric"] is metric_class


def test_register_duplicate_metric(metric_class):
    register_metric("TestMetric", metric_class)

    with pytest.raises(KeyError):
        register_metric("TestMetric", metric_class)


def test_unregister_metric(metric_class):
    register_metric("TestMetric", metric_class)

    assert Metrics()._metrics["TestMetric"] is metric_class
    unregister_metric("TestMetric")

    assert "TestMetric" not in Metrics()._metrics


def test_decorator():
    @custom_metric
    class TestMetric2(Metric):
        pass

    assert Metrics()._metrics["TestMetric2"] is TestMetric2


def test_binary_fbeta_metric():
    true = [
        [
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0]
        ],
        [
            [0, 1, 0],
            [1, 1, 1],
            [0, 1, 0]
        ],
        [
            [1, 1, 1],
            [0, 1, 0],
            [1, 1, 1]
        ],
        [
            [1, 0, 1],
            [0, 1, 0],
            [1, 0, 1]
        ]
    ]

    pred = [
        [
            [0, 0, 0],
            [0, 1, 0],
            [0, 1, 0]
        ],
        [
            [0, 1, 0],
            [1, 1, 1],
            [0, 0, 0]
        ],
        [
            [1, 1, 1],
            [0, 0, 1],
            [1, 1, 1]
        ],
        [
            [1, 1, 1],
            [0, 1, 0],
            [1, 0, 1]
        ]
    ]

    true = K.constant(true)
    pred = K.constant(pred)
    tp = 17
    fp = 2
    fn = 3

    eps = 1e-8

    def fscore(beta, tp, fp, fn):
        numerator = (1 + beta**2) * tp + eps
        denominator = (1 + beta**2) * tp + beta**2 * fn + fp + eps

        return numerator / denominator

    beta = 1
    metric = BinaryFbeta(beta=beta)
    assert np.isclose(K.eval(metric(true, pred)), fscore(beta, tp, fp, fn))

    beta = 2
    metric = BinaryFbeta(beta=beta)
    assert np.isclose(K.eval(metric(true, pred)), fscore(beta, tp, fp, fn))

    metric = BinaryFbeta(beta=beta)
    metric.update_state(true, pred)
    assert np.isclose(
        K.eval(metric.result()),
        fscore(beta, tp, fp, fn))


def test_fbeta_metric():
    true = [
        [
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0]
        ],
        [
            [0, 1, 0],
            [1, 1, 1],
            [0, 1, 0]
        ],
        [
            [1, 1, 1],
            [0, 1, 0],
            [1, 1, 1]
        ],
        [
            [1, 0, 1],
            [0, 1, 0],
            [1, 0, 1]
        ]
    ]

    pred = [
        [
            [0, 0, 0],
            [0, 1, 0],
            [0, 1, 0]
        ],
        [
            [0, 1, 0],
            [1, 1, 1],
            [0, 0, 0]
        ],
        [
            [1, 1, 1],
            [0, 0, 1],
            [1, 1, 1]
        ],
        [
            [1, 1, 1],
            [0, 1, 0],
            [1, 0, 1]
        ]
    ]

    true = K.constant(true)
    pred = K.constant(pred)

    values = [{'tp': 2,
               'fp': 0,
               'fn': 1},
              {'tp': 4,
               'fp': 0,
               'fn': 1},
              {'tp': 6,
               'fp': 1,
               'fn': 1},
              {'tp': 5,
               'fp': 1,
               'fn': 0}]

    eps = 1e-8

    def fscore(beta, values):
        res = 0
        for value in values:
            tp, fp, fn = value['tp'], value['fp'], value['fn']
            numerator = (1 + beta**2) * tp + eps
            denominator = (1 + beta**2) * tp + beta**2 * fn + fp + eps

            res += numerator / denominator

        return res / len(values)

    beta = 1
    metric = Fbeta(beta=beta)
    assert np.isclose(K.eval(metric(true, pred)), fscore(beta, values))

    beta = 2
    metric = Fbeta(beta=beta)
    assert np.isclose(K.eval(metric(true, pred)), fscore(beta, values))

    beta = 1
    metric = Fbeta(beta=beta)
    assert np.isclose(
        K.eval(metric(true, pred, sample_weight=[1, 1, 0, 0])),
        fscore(beta, values[:2]))

    beta = 2
    metric = Fbeta(beta=beta)
    assert np.isclose(
        K.eval(metric(true, pred, sample_weight=[1, 1, 2, 1])),
        fscore(beta, values + values[2:3]))

    metric = Fbeta(beta=beta)
    metric.update_state(true, pred, sample_weight=[1, 1, 2, 1])
    assert np.isclose(
        K.eval(metric.result()),
        fscore(beta, values + values[2:3]))


def test_dice_metric():
    true = [
        [
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0]
        ],
        [
            [0, 1, 0],
            [1, 1, 1],
            [0, 1, 0]
        ],
        [
            [1, 1, 1],
            [0, 1, 0],
            [1, 1, 1]
        ],
        [
            [1, 0, 1],
            [0, 1, 0],
            [1, 0, 1]
        ]
    ]

    pred = [
        [
            [0, 0, 0],
            [0, 1, 0],
            [0, 1, 0]
        ],
        [
            [0, 1, 0],
            [1, 1, 1],
            [0, 0, 0]
        ],
        [
            [1, 1, 1],
            [0, 0, 1],
            [1, 1, 1]
        ],
        [
            [1, 1, 1],
            [0, 1, 0],
            [1, 0, 1]
        ]
    ]

    true = K.constant(true)
    pred = K.constant(pred)

    values = [{'tp': 2,
               'fp': 0,
               'fn': 1},
              {'tp': 4,
               'fp': 0,
               'fn': 1},
              {'tp': 6,
               'fp': 1,
               'fn': 1},
              {'tp': 5,
               'fp': 1,
               'fn': 0}]

    eps = 1e-8

    def fscore(beta, values):
        res = 0
        for value in values:
            tp, fp, fn = value['tp'], value['fp'], value['fn']
            numerator = (1 + beta**2) * tp + eps
            denominator = (1 + beta**2) * tp + beta**2 * fn + fp + eps

            res += numerator / denominator

        return res / len(values)

    beta = 1
    metric = Dice(beta=beta)
    assert np.isclose(K.eval(metric(true, pred)), fscore(beta, values))

    beta = 2
    metric = Dice(beta=beta)
    assert np.isclose(K.eval(metric(true, pred)), fscore(beta, values))

    beta = 1
    metric = Dice(beta=beta)
    assert np.isclose(
        K.eval(metric(true, pred, sample_weight=[1, 1, 0, 0])),
        fscore(beta, values[:2]))

    beta = 2
    metric = Dice(beta=beta)
    assert np.isclose(
        K.eval(metric(true, pred, sample_weight=[1, 1, 2, 1])),
        fscore(beta, values + values[2:3]))

    metric = Dice(beta=beta)
    metric.update_state(true, pred, sample_weight=[1, 1, 2, 1])
    assert np.isclose(
        K.eval(metric.result()),
        fscore(beta, values + values[2:3]))



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.7"


from . import keras
from . import model
from . import data
from . import customize
from . import database
from . import automation



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


from ..utils import Singleton
from pymongo import MongoClient
from bson.objectid import ObjectId
from collections import OrderedDict
from datetime import datetime
from time import time
import json
import pandas as pd


class DBClient(metaclass=Singleton):  # pragma: no cover
    def __init__(self, *args, **kwargs):
        pass

    def create_table(self, table_name, *args, **kwargs):
        raise NotImplementedError

    def drop_table(self, table_name, *args, **kwargs):
        raise NotImplementedError

    def insert(self, table_name, items, *args, **kwargs):
        raise NotImplementedError

    def delete(self, table_name, *args, **kwargs):
        raise NotImplementedError

    def update(self, table_name, *args, **kwargs):
        raise NotImplementedError

    def update_insert(self, table_name, *args, **kwargs):
        raise NotImplementedError

    def find(self, table_name, *args, **kwargs):
        raise NotImplementedError

    def find_all(self, table_name, cols=None):
        raise NotImplementedError

    def find_by_col(self, table_name, col_name, val):
        raise NotImplementedError

    def find_by_id(self, table_name, id):
        raise NotImplementedError

    def find_max(self, table_name, query, col_name):
        raise NotImplementedError

    def find_min(self, table_name, query, col_name):
        raise NotImplementedError

    def to_pandas(self, val, *args, **kwargs):
        raise NotImplementedError

    def get_id(self, db_obj):
        raise NotImplementedError

    def df_to_json(self, df):
        cols = df.columns

        data = []
        for item in df.values:
            item_dict = {}
            for i in range(len(cols)):
                try:
                    json.dumps(item[i])
                    item_dict[cols[i]] = item[i]
                except Exception:
                    item_dict[cols[i]] = str(item[i])
            data.append(item_dict)
        return data


class MongoDBClient(DBClient, metaclass=Singleton):  # pragma: no cover
    def __init__(self, db_name, *args, **kwargs):
        self.client = MongoClient(*args, **kwargs)
        self.db = self.client[db_name]

    def create_table(self, table_name):
        return self.db[table_name]

    def drop_table(self, table_name, *args, **kwargs):
        return self.db[table_name].drop(*args, **kwargs)

    def insert(self, table_name, items, time_logs=False, *args, **kwargs):
        table = self.db[table_name]

        time_logs_obj = {}

        if time_logs:
            time_logs_obj = {
                "created": datetime.utcfromtimestamp(time())
            }

        if type(items) == list:
            return table.insert_many(
                [{**item, **time_logs_obj} for item in items], *args, **kwargs)
        elif type(items) == dict or type(items) == OrderedDict:
            return table.insert_one(
                {**items, **time_logs_obj}, *args, **kwargs)
        else:
            raise Warning("Failed to add items")
            return None

    def delete(self, table_name, query=None, items=None, *args, **kwargs):
        if items is None and query is None:
            raise Warning("No items or query is specified. No deletions")
            return None

        if query:
            table = self.db[table_name]
            return table.delete_many(query, *args, **kwargs)

        if items:
            if type(items) == dict:
                return table.delete_one({'_id', items['_id']}, *args, **kwargs)
            elif type(items) == list:
                expresssions = [{'_id': item['_id']} for item in items]
                return table.delete_many({
                    '$or': expresssions
                })
            else:
                raise Warning("No deletions")
                return None

    def update(self, table_name, query, new_val, time_logs=False,
               *args, **kwargs):
        table = self.db[table_name]

        time_logs_obj = {}

        if time_logs:
            time_logs_obj = {
                '$currentDate': {
                    'lastModified': True
                }
            }
        return table.update_many(
            query, {'$set': new_val, **time_logs_obj}, *args, **kwargs)

    def update_insert(self, table_name, query, new_val, *args, **kwargs):
        table = self.db[table_name]

        return table.update_one(query, {'$set': new_val}, upsert=True)

    def update_by_id(self, table_name, id, new_val, time_logs=False):
        table = self.db[table_name]

        time_logs_obj = {}

        if time_logs:
            time_logs_obj = {
                '$currentDate': {
                    'lastModified': True
                }
            }

        return table.update_one(
            {'_id': id}, {'$set': new_val, **time_logs_obj})

    def find(self, table_name, query=None, projection=None, *args, **kwargs):
        table = self.db[table_name]

        return table.find(query or {}, projection or {},
                          *args, **kwargs)

    def find_all(self, table_name, cols=None, id=True):
        table = self.db[table_name]
        if cols is None:
            if id:
                return table.find()
            else:
                return table.find({}, {'_id': 0})
        else:
            cols_dict = {col: 1 for col in cols}
            return table.find({}, {'_id': 1 if id else 0, **cols_dict})

    def find_by_col(self, table_name, col_name, val):
        table = self.db[table_name]

        if type(val) == list:
            expresssions = [{col_name: v} for v in val]
            return table.find({
                '$or': expresssions
            })

        return table.find({col_name: val})

    def find_by_id(self, table_name, id, custom_id=False):
        table = self.db[table_name]

        if type(id) == list:
            expresssions = [{'_id': self._serialize_id(id_) for id_ in id}]
            return table.find({
                '$or': expresssions
            })
        else:
            return table.find_one({'_id': self._serialize_id(id)})

    def _serialize_id(self, id, custom_id=False):
        if custom_id:
            return id
        if type(id) == str:
            return ObjectId(id)
        return id

    def to_fk(self, val):
        def str_to_id(v):
            if type(v) == str:
                return ObjectId(v)
            return v
        if type(val) == list:
            return [str_to_id(v) for v in val]
        if type(val) == str:
            return str_to_id(val)
        else:
            return val

    def find_max(self, table_name, query, col_name):
        table = self.db[table_name]

        res = table.find(query).sort(col_name, -1).limit(1)
        if res.count() > 0:
            return res[0]
        else:
            return None

    def find_min(self, table_name, query, col_name):
        table = self.db[table_name]

        res = table.find(query).sort(col_name, 1).limit(1)[0]

        if res.count() > 0:
            return res[0]
        else:
            return None

    def to_pandas(self, val, id=True):
        df = pd.DataFrame(list(val))

        # Delete _id
        if not id:
            del df['_id']

        return df

    def get_id(self, db_obj):
        return db_obj['_id']



import pytest
from deoxys.model import load_model, Model
import numpy as np
# import matplotlib.pyplot as plt
from tensorflow.keras.applications import vgg16
from tensorflow.keras.applications.vgg16 import preprocess_input
from tensorflow.keras.preprocessing import image


@pytest.fixture(scope='module', autouse=True)
def vgg():
    vgg = vgg16.VGG16(weights='imagenet', include_top=True)

    yield vgg


def deprocess_image(x):
    # normalize tensor: center on 0., ensure std is 0.1
    x -= x.mean()
    x /= (x.std() + 1e-5)
    x *= 0.1

    # clip to [0, 1]
    x += 0.5
    x = np.clip(x, 0, 1)

    # convert to RGB array
    x *= 255
    # x = x.transpose((1, 2, 0))
    x = np.clip(x, 0, 255).astype('uint8')
    return x


def load_image(path, target_size=(224, 224)):
    x = image.load_img(path, target_size=target_size)
    x = image.img_to_array(x)
    x = np.expand_dims(x, axis=0)
    x = vgg16.preprocess_input(x)

    return x


def test_activation_map(vgg):
    # load the vgg into the deoxys model
    deo = Model(vgg)

    # load an real image
    img = load_image('tests/img/cat.jpg')

    # View activation map, 1st filters
    layer_list = [layer.name for layer in vgg.layers[1:4]]

    for i, layer in enumerate(layer_list):
        outs = deprocess_image(
            deo.activation_map(layer, img)[0][..., 0])


def test_backprop(vgg):
    # load the vgg into the deoxys model
    deo = Model(vgg)

    # load an real image
    img = load_image('tests/img/cat.jpg')

    layer_list = [layer.name for layer in vgg.layers[1:4]]

    for i, layer in enumerate(layer_list):
        outs = deprocess_image(deo.backprop(layer, img)[0])

    deo.backprop(layer, img, mode='one')
    deo.backprop(layer, img, mode='all')


def test_deconvnet(vgg):

    # load the vgg into the deoxys model
    deo = Model(vgg)

    # load an real image
    img = load_image('tests/img/cat.jpg')

    layer_list = [layer.name for layer in vgg.layers[1:4]]

    for i, layer in enumerate(layer_list):
        deprocess_image(deo.deconv(layer, img, mode='min')[0])


def test_guided_backprop(vgg):
    # load the vgg into the deoxys model
    deo = Model(vgg)

    # load an real image
    img = load_image('tests/img/cat.jpg')

    layer_list = [layer.name for layer in vgg.layers[1:4]]

    for i, layer in enumerate(layer_list):
        outs = deprocess_image(deo.guided_backprop(layer, img, mode='mean')[0])


def test_activation_maximization(vgg):
    deo = Model(vgg)

    layer_list = [layer.name for layer in vgg.layers[1:4]]

    for i, layer in enumerate(layer_list):
        outs = deprocess_image(deo.activation_maximization(
            layer, epochs=5, step_size=2)[0])


def test_max_filter(vgg):
    deo = Model(vgg)

    img = load_image('tests/img/cat.jpg')

    layer_list = [layer.name for layer in vgg.layers[1:4]]

    for i, layer in enumerate(layer_list):
        deo.max_filter(layer, img)




import os


def read_file(filename):
    """
    Return the content of a file

    Parameters
    ----------
    filename : str
        file name

    Returns
    -------
    str
        content of the file
    """
    with open(filename, "r", encoding="utf-8") as f:
        return f.read()


def write_file(content, filename):
    """
    Write to a file

    Parameters
    ----------
    filename : str
        file name

    content: str
        content of the file
    """
    with open(filename, "w", encoding="utf-8") as f:
        f.write(content)


def write_byte(content, filename):
    """
    Write bytes to a file

    Write to a file

    Parameters
    ----------
    filename : byte
        file name

    content: str
        content of the file
    """
    with open(filename, "wb") as f:
        f.write(content)


def file_finder(filename, callback=None, **kwargs):   # pragma: no cover
    # check for file existence
    if os.path.isfile(filename):
        return filename
    # use callback to perform customized actions
    elif callable(callback):
        return callback(**kwargs)
    # default behavior
    else:
        # Choose between ignoring the file or enter a new filename
        msg = '"{}" not fould! Ignore?[y/n]: '.format(filename)
        if input(msg).upper() != 'Y':
            count = 0
            while not os.path.isfile(filename) and count < 3:
                msg = '"{}" not fould! Please enter the new filename: '.format(
                    filename)
                filename = input(msg)
                count += 1
            if count < 3:
                return filename
        print('File ignored!')

        return None



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest

from deoxys.model import Model, model_from_full_config
from deoxys.utils import read_file


def test_load():
    pass


def test_save():
    pass


def test_from_config():
    config = read_file('tests/json/sequential-config.json')
    model_from_full_config(config)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


from ..model.callbacks import callback_from_config
from ..utils import deep_copy


def load_train_params(train_params):
    """
    Create training parameters from configuration

    Parameters
    ----------
    train_params : dict
        dictionary of training parameters configuration (epochs, callbacks)

        Example:
        ```
        {
            "epoch": 10,
            "callbacks": [
                "CSVLogger",
                {
                    "class_name": "ModelCheckpoint"
                }
            ]
        }
        ```

    Returns
    -------
    dict
        Dictionary of training parameters objects
    """
    if not train_params:
        return {}
    params = deep_copy(train_params)
    for key, val in params.items():
        if key == 'callbacks':
            if type(val) == dict:
                params[key] = [callback_from_config(val)]
            if type(val) == list:
                callbacks = []
                for v in val:
                    if type(v) == dict:
                        callbacks.append(callback_from_config(v))
                    elif type(v) == str:
                        callbacks.append(
                            callback_from_config({'class_name': v}))
                params[key] = callbacks
    return params



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest
from deoxys.keras.layers import Layer as KerasLayer
from deoxys.model.layers import Layers
from deoxys.customize import register_layer, unregister_layer, custom_layer
from deoxys.utils import Singleton


@pytest.fixture(autouse=True)
def clear_singleton():
    Singleton._instances = {}  # clear singleton


@pytest.fixture
def layer_class():
    class TestLayer(KerasLayer):
        pass

    yield TestLayer


def test_is_singleton():
    layers_instance = Layers()
    another_instance = Layers()

    assert layers_instance is another_instance


def test_register_random_obj():
    with pytest.raises(ValueError):
        register_layer("TestLayer", object)


def test_register_layer_success(layer_class):
    register_layer("TestLayer", layer_class)

    assert Layers()._layers["TestLayer"] is layer_class


def test_register_duplicate_layer(layer_class):
    register_layer("TestLayer", layer_class)

    with pytest.raises(KeyError):
        register_layer("TestLayer", layer_class)


def test_unregister_layer(layer_class):
    register_layer("TestLayer", layer_class)

    assert Layers()._layers["TestLayer"] is layer_class
    unregister_layer("TestLayer")

    assert "TestLayer" not in Layers()._layers


def test_decorator():
    @custom_layer
    class TestLayer2(KerasLayer):
        pass

    assert Layers()._layers["TestLayer2"] is TestLayer2
